/*  $Id: CAction.hh,v 1.61 1999/07/16 20:18:49 shaggy Exp $
 *      Default C/C++ Parsing Actions
 *      relative path: .//lib/cparse
 */ 

#ifndef __CAction_hh__
#define __CAction_hh__

#include <CLex>
#include <vStrTab>
#include <vSymTab>
#include <vCAction>
#include <CRep>
#include <COut>


#define INTERNAL_ERROR  { internal_error(__FILE__, __LINE__, __FUNCTION__); \
			  for(;;); }

struct Param {
	type_t		type;		// type of parameter
	vStrTab::Id	name;		// formal name of parameter
	vCAction::Item	defval;		// default value
	Param		*next;	
	void clear() { memset(this, 0, sizeof *this); }
};

struct ParamList {
	Param	*head;
	uint8	nparam;
	bool	variable:1;
};

#define GEN_UITEM(t, n) \
	inline t const & n(vCAction::Item_ i) \
		{ return (t const &)i.val; } \
	inline t& n(vCAction::Item *i) \
		{ return (t&)i->val; }

// Things we wish were declared in the union in vCAction::Item,
// and inline conversions to get to those things:
struct ItemDecl {
	vStrTab::Id	id;
	Decl		*decl;
};
GEN_UITEM(ItemDecl, uDecl)
GEN_UITEM(ItemDecl, uValId)
GEN_UITEM(ItemDecl, uAggrId)
GEN_UITEM(ItemDecl, uEnumId)
GEN_UITEM(Expr, uExpr)
GEN_UITEM(ExprList, uExprList)
GEN_UITEM(Stmt, uStmt)
GEN_UITEM(StmtList, uStmtList)
struct ItemConvId {
	vStrTab::Id	id;
	type_t		type;
};
GEN_UITEM(ItemConvId, uConvId)
struct ItemOperId {
	vStrTab::Id	id;
};
GEN_UITEM(ItemOperId, uOperId)
struct ItemCtorId {
	vStrTab::Id	id;
	vStrTab::Id	aggrid;
};
GEN_UITEM(ItemCtorId, uCtorId)
GEN_UITEM(ItemCtorId, uDtorId)
GEN_UITEM(Param*, uParm)
GEN_UITEM(TDecl, uTypeList)
GEN_UITEM(TDecl, uScope)




class CAction : public vCAction {
protected:
	CLex &_clex;
private:
	vStrTab &_strtab;
	vSymTab &_global_scope;
public:
	typedef vStrTab::Id Id;

	CAction(CLex &clp, vStrTab &strtab, vSymTab &globlsym);
	void accept();
	void eol();
	csz* cstr(Token token) const;
	csz* cstr(vStrTab::Id id) const;
	sz * cstr(Item_ item, sz *buf, size_t buflen,
				bool show_type = false) const;
	Id id(Token, csz *s) { return _strtab.id(s); }
	Id id(csz *s) { return _strtab.id(s); }

	void eprintf(csz *fmt, ...);
	void mprintf(csz *prefix, csz *fmt, ...);
	void vmprintf(csz *prefix,
				csz *fmt, va_list args);
	void _assertfail(csz *cond, csz *filename,
				int line);

	void error(csz *fmt, ...);
	void warning(csz *fmt, ...);
	void abort(csz *fmt, ...);
	void internal_error(csz *file, int line,
				csz *module = 0);
	
	void mprint_action(csz *prefix, csz *name,
			Item_ p1, Item_ p2, Item_ p3);
	void mprint_result(csz *prefix, Item_ r);
	void unsupported(csz *name, Item_ p1, Item_ p2, Item_ p3);

	// vSymTab* get_symtab(Item_ id, Item_ scope);

	action Test, Undeclared;
	action Index, DoWhile;
	action Break, Continue;
	action NewAggr, NewNspace, NewLocal;
	action FreeScope;

	virtual void abort_exit() = 0;
	virtual void internal_error_exit() = 0;

private:
	int _inclevel;
	bool _cxx, _dcxx, _hdr;
protected:
	int _anon;
	bool _realtypes;
	bool _warnerr;
	bool _nowarn;
	bool _nooutput;
	Id _id_this;
	ScopeList _allscopes;
	Decl::OutType _otype;
public:
	CCOut ccout;
	bool _notedecl, _notedef;

	// Id string_id(Token, csz *str);
	// void abort_exit();
	// void internal_error_exit();
	// void veprintf(csz *fmt, va_list args);

	void realtypes(bool tf=true) { _realtypes = tf; }
	void warnerr(bool tf=true) { _warnerr = tf; }
	void nowarnings(bool tf=true) { _nowarn = tf; }
	void nooutput(bool tf) { _nooutput = tf; }
	bool nooutput() { return _nooutput || errors() ||
			   (_warnerr && warnings()); }
	Id new_anon(Scope, csz *tname);

	// void print_type(const type_t type);
	// sz* type_string(const type_t type, sz *s);

	type_t decl_type(Item_ item, bool impwarn = true);
	type_t decl_type(Item *r, Item_ item, bool impwarn = true);
	int type_sizeof(type_t type);
	int type_alignment(type_t type);

	Scope global() const { return &_global_scope; }

	void print_syms();
	void dump_scopes()
		{ ::dump_scopes(_allscopes); }

	Decl* forwref(TDecl::DataType type, Id id, Scope = 0, bool search = false);
	static Token decltoken(Decl *decl);

	Scope newscope(Scope parent, Decl *);
	Scope newaggrscope(Decl *);
	Scope setaggrscope(Decl *, Scope);
	Scope getaggrscope(Decl *);
	vSymTab* get_symtab(Decl *scope);
	vSymTab* get_symtab(Item_ scope);
	vSymTab* get_symtab(Item_ id, Item_ scope);

	action Literal, Primary;
	action UnaryOp, BinaryOp, CondExpr, OpNew;
	action DeclSym, DeclParam;
	action DeclPtr, DeclArray, DeclFunc;
	action SetAttr;
	action Define, DefParam;
	action BuildTypeList, BuildParamList;
	action AggrRef, AggrDecl, EnumRef;
	action NewScope, FuncScope, SetAccess, SetLinkage, Inherit;
	action EndAggr;
	action NewEnum, EnumNext;
	action FindSym, SpecialId, SpecialStr;

	action BuildExprList, BuildStmtList, BuildArray;
	action Cast, Call, CtorCall, Goto;
	action CtorInit;

	action IfElse, While, For, ForExpr, Switch;
	action Label, Return;

	long check_repeat(long a, long b, csz *name);

	void parse_dcxx()
		{ _hdr = false; _dcxx = _cxx = true; _clex.parse_dcxx(); ccout.hdr(_hdr); }
	void parse_dhxx()
		{ _hdr = _dcxx = _cxx = true; _clex.parse_dcxx(); ccout.hdr(_hdr); }
	void parse_cxx()
		{ _hdr = _dcxx = false; _cxx = true; _clex.parse_cxx(); ccout.hdr(_hdr); }
	void parse_hxx()
		{ _hdr = _cxx = true; _dcxx = false; _clex.parse_cxx(); ccout.hdr(_hdr); }
	void parse_c()
		{ _hdr = _dcxx = _cxx = false; _clex.parse_c(); ccout.hdr(_hdr); }
	void parse_h()
		{ _hdr = true; _dcxx = _cxx = false; _clex.parse_c(); ccout.hdr(_hdr); }
	bool dcxx() const { return _dcxx; }
	bool cxx() const { return _cxx; }
	bool hdr() const { return _hdr; }

	Decl* newdecl(Id id, Scope parent,
			TDecl::DataType = TDecl::dt_default, bool err = true);
	Decl* newdecl(Id id, Scope parent, type_t type,
			TDecl::DataType = TDecl::dt_default, bool err = true);
	// Decl* newdecl_item(Item_ type, Item_ sym, Item_ scope, bool err);
	// Decl* parentdecl(Decl *);
	// Decl* parentaggrdecl(Decl *);
	// void chaindecl(Decl *);
	// void unchaindecl(Decl *);
	csz* printable_scope_string(Scope);
	void error_id(csz *what, Id id, Scope, bool showtype);
	void error_dupdecl(Id id, Scope s)
		{ error_id("already declared", id, s, true); }
	void error_dupdef(Id id, Scope s)
		{ error_id("already defined", id, s, true); }
	void error_dupinit(Id id, Scope s)
		{ error_id("already initialized", id, s, false); }
	void error_notdecl(Id id, Scope s)
		{ error_id("not declared", id, s, false); }
	void error_notdef(Id id, Scope s)
		{ error_id("not defined", id, s, false); }
	void error_notinit(Id id, Scope s)
		{ error_id("not initialized", id, s, false); }
	void error_notfunc(Id id, Scope s)
		{ error_id("is not a function", id, s, false); }
	void error_shadows(Id id, Scope s)
		{ error_id("shadows parameter", id, s, false); }
	void error_noderef(Id id, Scope s)
		{ error_id("cannot be dereferenced", id, s, false); }
	void error_notaggr(Id id, Scope s)
		{ error_id("is not an aggregate", id, s, false); }
	csz* opname(Token t);
	bool makeconst(Item *r, Item_ arg);
	Expr exprof(Item_ item);
	Stmt stmtof(Item_ item);
	StmtList stmtlistof(Item_ item);
	Expr dotexpr(type_t t, Item_ m, type_t *rt = 0);
	Expr ptrexpr(type_t t, Item_ m, type_t *rt = 0);
	Decl* addconv(Scope, type_t, Decl*);
	Decl* addctor(Scope, Decl*);
	void inherit(Scope, Decl *base, Decl::Access acc = Decl::ac_public);
	void inherit(Decl *, Decl *base, Decl::Access acc = Decl::ac_public);

	void notesymdecl(Decl *decl);
	void notesymdef(Decl *decl);
	sz * scoped_string(Scope sc, sz *s)
		{ return ccout.scoped_string(sc, s); }
	sz * scoped_string(Decl *d, sz *s)
		{ return ccout.scoped_string(d, s, true); }
	sz * scoped_string_notag(Decl *d, sz *s)
		{ return ccout.scoped_string(d, s, false); }
	sz * type_string(type_t t, sz *s)
		{ return ccout.type_string(t, s); }
	Decl* addparam(int i, Scope, const Param &p);
	Decl* addthis(Decl *decl);
	Decl* getthis(Decl *decl) { return addthis(decl); }

	struct FnInfo {
		Func *f;
		Method *m;
		bool c;		// is const
		type_t o;	// obj type
		type_t r;	// return type
		type_t p;	// param type
		Scope sc;	// parameter scope
	};
	bool getfninfo(type_t, FnInfo*);
	Decl* findfn(Decl *df, ExprList arg, bool thisconst = false);
	Decl* findfn(Decl *df, const FnInfo*, FnInfo* = 0);
	Decl* findfn(Decl *df, type_t t);

	int _typematch(type_t to, type_t from);
	int typematch(type_t to, type_t from);
	int typematchctors(type_t to, type_t from);
	int fnmatch(type_t ft, ExprList arg, bool thisconst = false);

	type_t rtypeof(Item_ val);
	type_t deref(type_t type);

	void mergedefs(Func *to, Func *from);

	int inclevel() const { return _inclevel; }
	void inclevel(int i) { _inclevel = i; }
};

// XXX - probably ought to be static member fns:
type_t realtype(type_t ptype);
type_t convtype(type_t ptype);
TDecl::DataType datatypeof(vCAction::Token tok);
type_t type_ptrto(type_t t, bool kconst = false, bool kvolatile = false);
type_t type_refto(type_t t, bool kconst = false, bool kvolatile = false);
type_t type_qualifier(type_t t, bool kconst, bool kvolatile);
type_t type_const(type_t t);
type_t type_declref(Decl *d);
type_t type_array(type_t t, uint size, void *dsize = 0);
type_t type_aggr(vStrTab::Id id, Scope sc);
type_t type_enum(Decl *d);
type_t type_psig(uint np, const type_t *p, bool var = false);
type_t type_func(type_t rt, type_t p, vSymTab *s);
type_t type_method(type_t o, type_t f);
type_t realtype(type_t ptype);


#ifdef CACTION_DEBUG

extern bool debugaction;

#define PRINT_ACTION(r,p1,p2,p3) \
	PrintAction _p_a_ (this, __FUNCTION__, r, p1, p2, p3)

class PrintAction {
	CAction *_caction;
	CAction::Item *_rval;
public:
	PrintAction(CAction *caction, csz *name,
		CAction::Item *rval,
		CAction::Item_ p1,
		CAction::Item_ p2,
		CAction::Item_ p3)
			: _caction(caction), _rval(rval)
		{ if(debugaction)
			_caction->mprint_action("", name, p1, p2, p3);
		}
	~PrintAction()
		{ if(debugaction)
			_caction->mprint_result("", *_rval);
		}
};

#else
#define PRINT_ACTION(r,p1,p2,p3) 	
#endif


// CAction Item types (XXX - majorly out of date)
//	(see vCAction.hh for CParse Item types
//
//	AnyId:		any identifier (might not be declared)
//		= DeclId | NewId
//
//	AnyDeclId:	any declared indentifier
//		= AnyTypeId | ValId | NspaceId | TplateId
//
//	AnyTypeId:	identifier for any type descriptor
//		= AggrId | EnumId | TypeId
//
//	AnyValId:	identifier for any symbol with a value
//		= ValId
//	    destructor:
//		token= DTOR_ID
//		id= string id (~classname)
//		tuple.right= original string id (without the ~)
//	    operator:
//		token= OPER_ID
//		id= string id (full operator name)
//	    conversion:
//		token= CONV_ID
//		id= string id (full type name)
//		tuple.right= type ptr
//
//	AnyScopeId:	identifier for any scope
//		= AggrId | NspaceId
//
//	AggrId:		identifier for an aggregate (class/struct/union)
//		token= AGGR_ID
//		id= string id
//		tuple.right = Decl*
//
//	EnumId:		identifier for an enumeration
//		token= ENUM_ID
//		id= string id
//		tuple.right = Decl*
//
//	TypeId:		identifier for a type (not enum/aggr)
//		token= TYPE_ID
//		id= string id
//		tuple.right = Decl*
//
//	ValId:		identifier for a symbol with a value
//		token= VAL_ID
//		id= string id
//		tuple.right = Decl*
//
//	NewId:		undefined identifier
//		token= ID
//		id= string id
//
//	NspaceId:
//		token= NSPACE_ID | ASPACE_ID
//		id= string id
//
//	TypeList:	compound list of identifiers to describe a type
//		token= [TYPELIST]
//
//	Lit:		literal (and synthesized literals, which
//				are essentially constants)
//		token= [LITERAL] | NONE
//
//	Expr:		compound expression
//
//

#endif
