/*  $Id: CLex.hh,v 1.26 1999/07/16 20:18:49 shaggy Exp $
 *      C/C++ Lexical Analysis
 *      relative path: .//lib/cparse
 */ 

#ifndef __CLex_h__
#define __CLex_h__

#include <stdx1.h>
#include <Lexer>


class CLex : public Lexer {
	enum State {
		S_NEXT, S_SKIP, S_SKIP_NEXT, S_SPACE, S_SPACE_EOL,
		S_EOL, S_NUM, S_BIN, S_OCT, S_HEX,
		S_NUM_START, S_NUM_END, S_FLOAT, S_ID, S_SQUOTE,
		S_DQUOTE, S_DQUOTE2, S_BQUOTE, S_AQUOTE, S_QUOTE_ESC,
		S_QUOTE_OCT, S_QUOTE_HEX, S_OP_ADD, S_OP_SUB, S_OP_MUL,
		S_OP_DIV, S_OP_MOD, S_OP_AND, S_OP_OR, S_OP_XOR,
		S_OP_LNOT, S_OP_ASN, S_OP_LSS, S_OP_GTR, S_OP_LSHIFT,
		S_OP_RSHIFT, S_COLON, S_DOT, S_DOT2, S_PTRTO,
		S_QMARK, S_CPP, S_NOCPP, S_CPP_NUM, S_CPP_NONUM,
		S_CPP_NAME, S_CPP_NONAME, S_CPP_OPTNUM,S_CPP_NOOPTNUM,S_CPP_CMD,
		S_CPP_ARG1, S_COMMENT, S_COMMENT2, S_EAT, S_EAT_STR,
		S_LINE, S_LINE_STR
	} state;
	enum State old_state;
	char *str;		// string to store tokens in
	char *strp;		// current pointer in 'str'
	size_t str_size;	// total size of 'str'
	char *buf;		// buffer to store translations in
	char *bufp;		// current pointer in 'buf'
	size_t buf_size;	// total size of 'buf'
	char *name;		// buffer to store source name in
	char *namep;		// current pointer in 'name'
	size_t name_size;	// total size of 'name'
	uint64 num;	// numeric value of last token
	double fnum;	// floating point value of last token
	size_t size;	// size of last token
	bool _ws;	// enable whitespace recognition
	bool _indent;	// enable indent whitespace recognition
	bool _kw;	// enable keyword lookup
	bool _cpp;	// return cpp tokens
	bool _cxx;	// parse C++ tokens and keywords
	bool _dcxx;	// parse DC++ tokens and keywords
	bool _eol;	// return EOL tokens
	bool _in_cpp;	// currently in the middle of a cpp line
	bool _qwrap;	// wrap quotes until complete
	bool _cpp_first; // first argument after '#'
	bool _first;	// first token in line
	bool _signed;	// last num was signed (unused)
	int _long;	// last num was long (unused)
public:
	enum Token {
		NONE = 0,
		EOF_COMPLETE,
		EOF_INCOMPLETE,
		INCOMPLETE, INVALID,

#include "CParse.tab.h"

		NUM_TOKENS,

		START_CPP = CPP_DEFINE,
		END_CPP = CPP__STDCVER,
		START_C = KW_ATTRIBUTE,
		END_C = KW__PFUNC,
		START_CXX = KW_OPERATOR,
		END_CXX = OP_DELETE,
		START_DCXX = KW_DYNAMIC,
		END_DCXX = KW_DYNAMIC
	};
private:
	static csz* token_table[];
	int _include_level;
	Token token;
public:
	CLex(char *buf, size_t str_size);
	CLex& set_buf(char *ptr, size_t size)
		{ buf = ptr; buf_size = size; return *this; }
	CLex& set_tok(char *ptr, size_t size)
		{ str = ptr; str_size = size; return *this; }
	CLex& set_name(char *ptr, size_t size)
		{ name = ptr; name_size = size; return *this; }
	CLex& parse_space(bool tf = true) { _ws = tf; return *this; }
	CLex& parse_indent(bool tf = true) { _indent = tf; return *this; }
	CLex& parse_keywords(bool tf = true) { _kw = tf; return *this; }
	CLex& parse_c() { _cxx = _dcxx = false; return *this; }
	CLex& parse_cpp(bool tf = true) { _cpp = tf; return *this; }
	CLex& parse_cxx() { _cxx = true; _dcxx = false; return *this; }
	CLex& parse_dcxx() { _dcxx = true; _cxx = true; return *this; }
	CLex& parse_eol(bool tf = true) { _eol = tf; return *this; }
	void get_line() { strp = str; state = S_LINE; }
	Token next_token();
	csz* token_string();
	static csz* token_string(Token t);
	static csz* token_string(Token t, bool tf);
	csz* token_string(Token t, uint64 val,
			bool is_signed = true, int is_long = 0);
	csz* token_string(Token t, csz *str, size_t len = NPOS);
	static csz* token_spacing(Token prev, Token next);
	Token last_token() const { return token; }
	uint64 last_num() const { return num; }
	uint last_char() const;
	bool last_bool() const { return (bool)num; }
	bool last_signed() const { return _signed; }
	int last_long() const { return _long; }
	double last_float() const { return fnum; }
	csz* last_str() const { return str; }
	csz* last_id() const { return str; }
	size_t last_size() const { return size; }
	csz* source() const { return name; }
	void source(csz* str) { strcpy(name, str); }
	int include_level() { return _include_level; }
	bool in_string() const { return state == S_DQUOTE; }
	bool in_char() const { return state == S_SQUOTE; }
	bool in_bquote() const { return state == S_BQUOTE; }
	void set_token(Token t) { token = t; }
};

#endif
