/*  $Id: COut.hh,v 1.17 1999/07/16 20:18:49 shaggy Exp $
 *      C++ output
 *      relative path: ./lib/cparse
 */ 

#ifndef __COut_hh__
#define __COut_hh__

#include <CObj>
#include <CLex>
#include <vStrTab>
#include <vSymTab>
#include <CRep>

class TokenList;

class CCOut {
	const Scope _scope;
	const vStrTab *_strtab;
	TDecl::Access _access;
	vErrorReport &_err;
	char _buf[65536];
	CLex _clex;
	Decl *_curdecl;		// for error reporting
	int _nl;
	int _x;
	int _indent;
	int _tabwid;
	bool _ws;
	char *_obuf;
	bool _fullscope;
	bool _debugtypes;
	bool _debugexpr;
	int _wid;
	int _outmask;
	csz *_sectionlabel;
	bool _interface;
	bool _onlyinterface;
	bool _hdr;

	csz* cstr(vStrTab::Id id) { return _strtab->string(id); }
	void printliteral(type_t type, uint64 val);
	// void printlist(ExprListNode *n);
	Decl* has_dependency(Decl *decl, bool needdef);
	// static bool has_dependencies(type_t type, bool needdef);
	void print_storageclass(Decl *decl);
	void regurgitate(TokenList *t);
	void print_access(Decl *decl);
	void print_otype(Decl *decl);
	void define_sym(Decl *decl, bool eos = true);
	void define_enum(Decl *decl);
	void define_aggr(Decl *decl);
	void declare_sym(Decl *decl, bool eos = true);
	// void print_types(csz *prefix, const Scope);
	// void print_fndecls(csz *prefix, const Scope, bool exp, bool stat);
	// void print_vardecls(csz *prefix, const Scope, bool exp);
	void _assertfail(csz *cond, csz *filename, int line);
public:
	enum DeclTypeMask {
		DTM_FN=		0x01,	// function
		DTM_TYPE= 	0x02,	// type
		DTM_VAR= 	0x04,	// variable
		DTM_CONST=	0x08,	// constant
		DTM_EXP= 	0x10,	// export
		DTM_IMP= 	0x20,	// import
		DTM_LOC= 	0x40,	// local
	};

	CCOut(vErrorReport &err, const Scope,
		TDecl::Access access = TDecl::ac_public);
	void debugtypes(bool d = true) { _debugtypes = d; }
	void debugexpr(bool d = true) { _debugexpr = d; }
	void printexpr(Expr, bool paren = false);
	void printstmt(Stmt);
	void printstmti(Stmt);
	void printexprlist(ExprList);
	void printstmtlist(StmtList);
	void printdecllist(StmtList);
	void printctorinit(StmtList);
	sz* literal_string(type_t type, uint64 val, sz *s);
	sz* type_string(type_t type, sz *s);
	sz* dbg_type_string(type_t type, sz *s);
	bool print_decls(const Scope sc, bool checkdep, bool err, int dtmask);
	void print_decls(const Scope, TDecl::Access acc = TDecl::ac_public);
	bool print_types(const Scope sc)
		{ return print_decls(sc, true, false, DTM_TYPE); }
	void print_fn_imports(const Scope sc)
		{ print_decls(sc, false, true, DTM_FN|DTM_IMP); }
	void print_fn_exports(const Scope sc)
		{ print_decls(sc, false, true, DTM_FN|DTM_EXP); }
	void print_fn_local(const Scope sc)
		{ print_decls(sc, false, true, DTM_FN|DTM_LOC); }
	bool print_constants(const Scope sc)
		{ return print_decls(sc, true, false, DTM_CONST); }
	void print_var_imports(const Scope sc)
		{ print_decls(sc, false, true, DTM_VAR|DTM_IMP); }
	void print_var_exports(const Scope sc)
		{ print_decls(sc, false, true, DTM_VAR|DTM_EXP); }
	void print_defs(const Scope);
	void print_reqs(const Scope);
	sz *scoped_string(const Scope, sz *s);
	sz *scoped_string(Decl *decl, sz *s, bool needtag = false);
	sz *fullscoped_string(Decl *decl, sz *s);
	void fullscope(bool tf = true) { _fullscope = tf; }
	void onl(int y = 1);
	void onltab(int x = -1);
	void rshift()  { _indent += _tabwid; }
	void lshift()  { _indent -= _tabwid; }
	void ws() { _ws = true; }
	void sectionlabel(csz *label) { _sectionlabel = label; }
	int oprintf(csz *fmt, ...);
	void width(int w) { _wid = w; }
	int width() const { return _wid; }
	void outmask(int m) { _outmask = m; }
	int outmask() const { return _outmask; }
	Scope scope() const { return _scope; }
	void scope(Scope s) { _scope = s; }
	void interface(bool tf) { _interface = tf; }
	void onlyinterface(bool tf) { _onlyinterface = tf; }
	bool interface() const { return _interface || _onlyinterface; }
	bool onlyinterface() const { return _onlyinterface; }
	void hdr(bool tf) { _hdr = tf; }
	bool hdr() const { return _hdr; }
};


sz *strpcat(sz *str, csz *prefix);
csz *simple_typename(type_t type);

#endif
