/*  $Id: CPPParse.y,v 1.5 1999/06/20 17:10:35 shaggy Exp $
 *      C/C++ Pre-Processor Parser (yacc grammar)
 *      
 *      relative path: lib/cparse
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

%pure_parser

%{

#pragma Make Includes_Banned += -nostdinc -nostdinc++

#include <CPPParse>

// #define YYSTYPE CAction::Item

#define yyparse CPPParse::parse
#define yylex caction->lex
#define yyerror caction->error
#define yywarning caction->warning


#define do(r,a)		caction->action(r,CAction::a,r,r,r,r)
#define do1(r,a,p1)	caction->action(r,CAction::a,p1,r,r,r)
#define do2(r,a,p1,p2)	caction->action(r,CAction::a,p1,p2,r,r)
#define do3(r,a,p1,p2,p3)	caction->action(r,CAction::a,p1,p2,p3,r)
#define do4(r,a,p1,p2,p3,p4)	caction->action(r,CAction::a,p1,p2,p3,p4)
#define change(r,t)	(r.token = CLex::t, r)

%}


%token _EMPTY		// used for precedence


// Generic tokens:

%token _ID		// identifier
%token _NUM		// integer-literal
%token _CHAR		// character-literal
%token _FLOAT		// floating-literal
%token _STR		// string-literal
%token _BOOL		// boolean-literal
%token _BACKSTR		// `string`
%token _ANGSTR		// <string>
%token _LINE		// % line
%token _SPACE		// whitespace
%token _EOL		// end of line

// CPP tokens:

%token _CPP		// #
%token _CPP_JOIN	// ##
%token _CPP_DEFINED 	// "defined"

// CPP keywords:

%token _CPP_DEFINE	// "define"
%token _CPP_UNDEF	// "undef"
%token _CPP_IF		// "if"
%token _CPP_IFDEF	// "ifdef"
%token _CPP_IFNDEF	// "ifndef"
%token _CPP_ENDIF	// "endif"
%token _CPP_ELIF	// "elif"
%token _CPP_INCLUDE	// "include"
%token _CPP_PRAGMA	// "pragma"
%token _CPP_ASSERT	// "assert"
%token _CPP_UNASSERT	// "unassert"
%token _CPP_LINE	// "line"
%token _CPP_ERROR	// "error"
%token _CPP_WARNING	// "warning"
%token _CPP_IDENT	// "ident"

// standard C punctuation:

%token _EOS _COLON _COMMA		// ; : ,
%token _QMARK _POUND _LBRACE _RBRACE	// ? # { }
%token _LBRACK _RBRACK _LPAREN _RPAREN	// [ ] ( )

// standard C operators:

%token _OP_ADD _OP_SUB			// + -
%token _OP_MUL _OP_DIV _OP_MOD		// * / %
%token _OP_NOT _OP_AND _OP_OR _OP_XOR	// ~ & | ^
%token _OP_LNOT _OP_LAND _OP_LOR _OP_LXOR // ! && || ^^
%token _OP_LSS _OP_GTR _OP_LEQ _OP_GEQ	// < > <= >=
%token _OP_EQU _OP_NEQ _OP_INC _OP_DEC	// == != ++ --
%token _OP_LSHIFT _OP_RSHIFT		// << >>
%token _OP_ASN				// =
%token _OP_ASNADD _OP_ASNSUB		// += -=
%token _OP_ASNMUL _OP_ASNDIV _OP_ASNMOD	// *= /= %=
%token _OP_ASNAND _OP_ASNOR _OP_ASNXOR	// &= |= ^=
%token _OP_ASNLSHIFT _OP_ASNRSHIFT	// <<= >>=
%token _OP_DOT _OP_PTR			// . ->

// unary C operators not generated by lexer:

%token _OP_REF				// &
%token _OP_DEREF			// *
%token _OP_UMINUS			// -
%token _OP_INCP				// ++
%token _OP_DECP				// --


// precedence rules:

%left _EMPTY
%left error

%left _ID
%left _LBRACE _COMMA _EOS

%nonassoc _KW_THROW

%right _COLON
%right _OP_ASN _OP_ASNADD _OP_ASNSUB _OP_ASNMUL _OP_ASNDIV _OP_ASNMOD _OP_ASNAND _OP_ASNOR _OP_ASNXOR _OP_ASNRSHIFT _OP_ASNLSHIFT
%right _QMARK
%left _OP_LOR
%left _OP_LXOR
%left _OP_LAND
%left _OP_OR
%left _OP_XOR
%left _OP_AND
%left _OP_EQU _OP_NEQ
%left _OP_LEQ _OP_GEQ _OP_LSS _OP_GTR
%left _OP_LSHIFT _OP_RSHIFT
%left _OP_ADD _OP_SUB
%left _OP_MUL _OP_DIV _OP_MOD
%left _OP_MPTR _OP_MDOT
%left _OP_UMINUS _OP_UPLUS _OP_REF _OP_DEREF _OP_NOT _OP_INC _OP_DEC _OP_INCP _OP_DECP _OP_LNOT
%left _OP_PTR _OP_DOT _LPAREN _LBRACK

%%

// For the syntax descriptions,
// the section numbers refer to the April 28, 1995 version of 
// the "Working Paper for Draft Proposed International Standard
// for Information Systems- Programming Language C++"

// unquoted names represent other syntactical parts
// (which are italicized in the document), and quoted names
// represent literal text (keywords)
// any item followed by "(opt)" is optional.

// A lot of this is really ugly, but unfortunately most
// of the parser had to be pre-chewed to keep yacc
// from choking on it
// This involved taking major sections of the BNF
// and flattening them, causing many redundant
// definitions that look very similiar with minor
// changes for all the optional parts
// A lot of this is the result of the fact that
// class names can also be used as scope specifiers


// preprocessing-file: (s16)
//	group(opt)

// group: (s16)
//	group-part
//	group group-part

// group-part: (s16)
//	pp-tokens(opt) new-line
//	if-section
//	control-line

// if-section: (s16)
//	if-group elif-groups(opt) else-group(opt) endif-line

// if-group: (s16)
//	# "if"		constant-expression new-line group(opt)
//	# "ifdef"	identifier new-line group(opt)
//	# "ifndef"	identifier new-line group(opt)

// elif-groups: (s16)
//	elif-group
//	elif-groups elif-group

// elif-group: (s16)
//	# "elif"	constant-expression new-line group(opt)

// else-group (s16)
//	# "else"	new-line group(opt)

// endif-line: (s16)
//	# "endif"	new-line

// control-line: (s16)
//	# "include"	pp-tokens new-line
//	# "define"	identifier replacement-list new-line
//	# "define"	identifier lparen identifier-list(opt) ) 
//				replacement-list new-line
//	# "undef"	identifier new-line
//	# "line"	pp-tokens new-line
//	# "error"	pp-tokens(opt) new-line
//	# "pragma"	pp-tokens(opt) new-line
//	#		new-line

// lparen: (s16)
//	the left-parenthesis character without preceeding white-space

// replacement-list: (s16)
//	pp-tokens(opt)

// pp-tokens: (s16)
//	preprocessing-token
//	pp-tokens preprocessing-token

// new-line: (s16)
//	the new-line character


%%

