/*  $Id: CRep.cc,v 1.25 1999/07/16 20:18:49 shaggy Exp $
 *      C++ internals (knowledge representation and manipulation)
 *      
 *      relative path: lib/cparse
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#include <libc.h>
#include <stdx1.h>
#include <defines>
#include <ldefines>
#include <StrTab>
#include <SymTab>

#include "CRep.hh"


// lexical shortcuts
typedef vStrTab::Id Id;
typedef vCAction::Token Token;


csz* DeclCommon::_scname[] = {"default", "auto", "static", "extern",
			"register", "typedef", "mutable", "constant",
			"param", "friend", "extern \"C\"", "extern \"C++\"",
			"?", "?", "?", "?"};

csz* DeclCommon::_dtname[] = {"default", "int", "char", "wchar",
			"bool", "void", "float", "double",
			"enum", "struct", "class", "union",
			"namespace", "enumval", "type", "dtype",
			"dclass", "impl"};

uint8 DeclCommon::_dtflags[] = {
	/* default */	DTF_SIGNOK|DTF_SHORTOK|DTF_LONGOK,
	/* int */	DTF_SIGNOK|DTF_SHORTOK|DTF_LONGOK,
	/* char */	DTF_SIGNOK,
	/* wchar */	0,
	/* bool */	0,
	/* void */	0,
	/* float */	0,
	/* double */	DTF_LONGOK,
	/* enum */	DTF_TAG,
	/* struct */	DTF_TAG|DTF_AGGR|DTF_SCOPE,
	/* class */	DTF_TAG|DTF_AGGR|DTF_SCOPE,
	/* union */	DTF_TAG|DTF_AGGR|DTF_SCOPE,
	/* namespace */	DTF_TAG|DTF_SCOPE,
	/* enumval */	0,
	/* type */	DTF_TAG|DTF_AGGR|DTF_SCOPE,
	/* dtype */	DTF_TAG|DTF_AGGR|DTF_SCOPE,
	/* dclass */	DTF_TAG|DTF_AGGR|DTF_SCOPE,
	/* impl */	DTF_AGGR|DTF_SCOPE,
};

csz* DeclCommon::_acname[] = {"public", "protected", "private", "?"};

csz* Decl::_otname[] = { "noout", "fromhdr", "hdr", "code", "repos" };

obj_t Type_DeclRef;
obj_t Type_Enum;


Decl*
symget(Id id, const Scope sc, bool tagged)
{
	for(Decl *d = sc.get(id); d; d = d->link()) {
		if(!tagged || d->tagged())
			return d;
	}
	for(Decl *a = sc.get(ID_ANON); a; a = a->link()) {
		if(DeclRef *dref = dcast_DeclRef(a->ptype())) 
			if(Aggr *aggr = dcast_Aggr(dref->decl()->ptype())) 
				if(Decl *d = symget(id, aggr->scope(), tagged))
					return d;
	}
	for(Decl *i = sc.get(ID_INHERIT); i; i = i->link()) {
		if(Decl *d = symget(id, i->scope(), tagged))
			return d;
	}
	return 0;
}

Decl*
symfind(Id id, const Scope sc, bool tagged, bool search)
{
	Decl *d;
	if(id == sc.pchid())
		sc = sc.parent();
	do {
		if((d = symget(id, sc, tagged))) 
			return d;
		sc = sc.parent();
	} while(search && sc);
	return 0;
}


Expr
Expr::dup() const
{
	ExprNode *n = new ExprNode;
	*n = *p;
	return Expr(n);
}

Expr::Expr(type_t t)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(TYPE);
	p->rtype = t;
}

Expr::Expr(const Lit &lit)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(LITERAL);
	p->lit = lit;
	p->rtype = lit.type;
}

Expr::Expr(Decl *decl)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(VAL_ID);
	p->decl = decl;
	p->rtype = decl->rtype();
}

Expr::Expr(TokenList *tlist)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(TOKENLIST);
	p->tlist = tlist;
	p->rtype = 0;
}

Expr::Expr(Id id)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(ID);
	p->id = id;
	p->rtype = 0;
}

Expr::Expr(Token tok, Expr e)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->rtype = e.p ? e.p->rtype : 0;
	switch(tok) {
	case T(OP_INCP):
	case T(OP_DECP):
		p->e.left = e.p;
		p->e.mid = 0;
		p->e.right = 0;
		break;
	default:
		p->e.left = 0;
		p->e.mid = 0;
		p->e.right = e.p;
	}
}

Expr::Expr(Token tok, Expr e1, Expr e2)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = 0;
	p->e.right = e2.p;
	p->rtype = e1.p ? e1.p->rtype : 0;
}

Expr::Expr(Token tok, Expr e1, Expr e2, Expr e3)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = e2.p;
	p->e.right = e3.p;
	p->rtype = e3.p ? e3.p->rtype : 0;
}


bool
Expr::operator==(Expr e) const
{
	if(op() != e.op())
		return false;
	if(rtype() != e.rtype())
		return false;
	switch(op()) {
	case T(LITERAL):
		return lit() == e.lit();
	default:
		return left() == e.left() 
			&& mid() == e.mid()
			&& right() == e.right();
	}
}


ExprList::ExprList(Token tok, Token fmt)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.tok_mid = fmt;
	p->e.right = 0;
	p->rtype = 0;
}

ExprList::ExprList(Token tok)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.tok_mid = T(NONE);
	p->e.right = 0;
	p->rtype = 0;
} 

ExprList&
ExprList::append(Expr e)
{
	e.p->next = 0;
	if(p->e.right)
		p->e.right->next = e.p;
	else
		p->e.left = e.p;
	p->e.right = e.p;
	return *this;
}

// XXX - currently untested
ExprList&
ExprList::append(ExprList e)
{
	if(e.p) {
		if(p->e.right)
			p->e.right->next = e.p;
		else
			p->e.left = e.p;
		ExprNode *p;
		for(p = e.p; ExprNode *n = p->next; p = n)
			;
		p->e.right = p;
	}
	return *this;
}

int
ExprList::len() const
{
	int n;
	Expr e = first();
	for(n = 0; e; e = e.next())
		n++;
	return n;
}

Stmt::Stmt(const Lit &lit)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(LITERAL);
	p->lit = lit;
	p->rtype = 0;
}


Stmt::Stmt(TokenList *tlist)
{
	p = new ExprNode;
	p->next = 0;
	p->op = T(TOKENLIST);
	p->tlist = tlist;
	p->rtype = 0;
}

Stmt::Stmt(Token tok, Expr e)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.mid = 0;
	p->e.right = e.p;
	p->rtype = 0;
}

Stmt::Stmt(Token tok, Expr e1, Expr e2)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = 0;
	p->e.right = e2.p;
	p->rtype = 0;
}

Stmt::Stmt(Token tok, Expr e1, Expr e2, Expr e3)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = e1.p;
	p->e.mid = e2.p;
	p->e.right = e3.p;
	p->rtype = 0;
}

StmtList::StmtList(Token tok, Scope scope)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.sc_mid = scope;
	p->e.right = 0;
	p->rtype = 0;
}

StmtList::StmtList(Scope scope, Token tok)
{
	p = new ExprNode;
	p->next = 0;
	p->op = tok;
	p->e.left = 0;
	p->e.sc_mid = scope;
	p->e.right = 0;
	p->rtype = 0;
}

StmtList&
StmtList::append(Stmt s)
{
	if(p->e.right)
		p->e.right->next = s.p;
	else
		p->e.left = s.p;
	p->e.right = s.p;
	return *this;
}

int
tokenprec(Token tok, int assoc)
{
	switch(tok) {
	case T(TOKENLIST):
	case T(LITERAL):
	case T(VAL_ID):
		return 100;
	case T(OP_DOT): case T(OP_PTRTO): case T(OP_INDEX):
	case T(KW_SCAST): case T(KW_DCAST): case T(KW_CCAST): case T(KW_RCAST):
	case T(OP_FUNC):
		return 76;
	case T(OP_INCP): case T(OP_DECP):
		return 76+assoc;
	case T(OP_SIZEOF):
	case T(OP_INC): case T(OP_DEC):	
	case T(OP_NOT): case T(OP_LNOT):
	case T(OP_UMINUS): case T(OP_UPLUS):
	case T(OP_REF): case T(OP_PTR):
	case T(OP_NEW): case T(OP_DELETE):
	case T(OP_NEWA): case T(OP_DELETEA):
		return 72+assoc;
	case T(OP_MDOT): case T(OP_MPTRTO):
		return 68;
	case T(OP_MUL): case T(OP_DIV): case T(OP_MOD): 
		return 64;
	case T(OP_ADD): case T(OP_SUB):
		return 60;
	case T(OP_LSHIFT): case T(OP_RSHIFT):
		return 56;
	case T(OP_LSS): case T(OP_LEQ): case T(OP_GTR): case T(OP_GEQ):
		return 52;
	case T(OP_EQU): case T(OP_NEQ):
		return 48;
	case T(OP_AND):
		return 44;
	case T(OP_XOR):
		return 40;
	case T(OP_OR):
		return 36;
	case T(OP_LAND):
		return 32;
	case T(OP_LXOR):
		return 28;
	case T(OP_LOR):
		return 24;
	case T(OP_ASN):
	case T(OP_ASNMUL): case T(OP_ASNDIV): case T(OP_ASNMOD):
	case T(OP_ASNADD): case T(OP_ASNSUB): 
	case T(OP_ASNLSHIFT): case T(OP_ASNRSHIFT):
	case T(OP_ASNAND): case T(OP_ASNOR): case T(OP_ASNXOR):
	case T(OP_CONDASN):
		return 20;
	case T(QMARK):
		return 16;
	case T(KW_THROW):
		return 12+assoc;
	case T(OP_COMMA):
		return 8;
	default:
		return 0;
	}
}

bool
scope_empty(const Scope sc)
{
	Id id;
	for(int n=0; (id = sc.index(n)); n++) 
		if(id != ID_INHERIT)
			return false;
	return true;
}

bool
scope_derived(const Scope sc)
{
	return (sc.get(ID_INHERIT)) ? true : false;
}

Id
Scope::pchid() const
{
	Decl *d = decl();
	return d ? d->id() : 0;
}


static void
dumpdecl(Decl *d)
{
	printf("%p(%s,%s,%s,%s%s%s%s%s%s%s%s%s%s) ->%p ", d,
		d->acname(), d->scname(), d->dtname(), d->otname(),
		d->defined() ? ",D" : "",
		d->initialized() ? ",I" : "",
		d->condasn() ? ",c" : "",
		d->anon() ? ",A" : "",
		d->field() ? ",F" : "",
		d->bitfield() ? ",B" : "",
		d->conv() ? ",C" : "",
		d->isfunc() ? ",f" : "",
		d->ismeth() ? ",m" : "",
		d->next());
	if(Decl *c = d->conv())
		printf("(c=%p)", c);
}

void
Expr::dump(int indent, csz *prefix) const
{
	if(!p) 
		return;
	printf("%*s%s", indent, "", prefix);
	printf("%p: %s (%d) ", p, CLex::token_string((CLex::Token)op()), op());
	switch(op()) {
	case T(TYPE):
		printf("t=%p\n", p->rtype);
		break;
	case T(LITERAL):
		printf("%p: %p: %lld\n", p->rtype, p->lit.type, p->lit.val);
		break;
	case T(VAL_ID):
		printf("t=%p d=%p %s: ", p->rtype, p->decl, p->decl->cstr_id());
		dumpdecl(p->decl);
		printf("\n");
		break;
	case T(TOKENLIST):
		printf("tl=%p\n", p->tlist);
		break;
	case T(ID):
		printf("%p\n", p->id);
		break;
	case T(STMTLIST):
	case T(EXPRLIST):
	case T(ARRAY):
		printf("m=%p r=%p\n", p->e.mid, p->e.right);
	 	left().dump(indent+4);
		break;
	case T(DECL):
		printf("\n");
	 	left().dump(indent+4);
		if(left().op() == T(VAL_ID)) {
			Decl *d = left().decl();
			if(d->initialized())
				d->vexpr().dump(indent+8);
		}
		break;
	default:
		printf("\n");
	 	left().dump(indent+4, "l: ");
	 	mid().dump(indent+4, "m: ");
	 	right().dump(indent+4, "r: ");
	}
	if(next()) 
		next().dump(indent);
}

void
dump_scope(const Scope scope)
{
	vStrTab *strtab = scope.strtab();
	Scope parent = scope.parent();
	printf("Scope %p: %s (%p: %s) ", (vSymTab*)(scope), scope.pchid()
			? strtab->string(scope.pchid()) : "?",
			(vSymTab*)(parent), parent ? (parent.pchid()
			? strtab->string(parent.pchid()) : "?") : "?");
	if(scope.decl())
		printf("(%p: a=%d,%p)", scope.decl(),
					scope.decl()->asize(),
					scope.decl()->afirst());
	printf("\n");
	Id id;
	for(int n=0; (id = scope.index(n)); n++) {
		csz *s = "?";
		if(id == ID_INHERIT)
			s = "<inherit>";
		else if(id == ID_USING)
			s = "<using>";
		else if(id == ID_ANON)
			s = "<anon>";
		else if(id == ID_FRIEND)
			s = "<friend>";
		else if(id == ID_CONV)
			s = "<conv>";
		else if(id == ID_CTOR)
			s = "<ctor>";
		else if(id)
			s = strtab->string(id);
		printf("  %s: ", s);
		Decl *d = scope.get(id);
		while(d) {
			dumpdecl(d);
			if(id == ID_INHERIT)
				printf("(i=%p)", d->idecl());
			if(d->dtaggr())
				printf("(a=%d,%p)", d->asize(), d->afirst());
			d = d->link();
			printf(d ? "\n     " : "\n");
		}
	}
}

void
dump_scopes(ScopeList slist)
{
	for( ;Scope s = slist.head(); slist = slist.tail()) 
		dump_scope(s);
}

Decl *
param_decl(Func *func, int i)
{
	Scope scope = func->scope();
	if(!scope)
		return 0;
	Id id;
	for(int n=0; (id = scope.index(n)); n++) {
		for(Decl *d = scope.get(id); d; d = d->link())
			if(d->storage() == Decl::sc_param && d->where() == i)
				return d;
	}
	return 0;
}


bool
subscope(Scope scope, Scope sub)
{
	if(scope) {
		if(sub == scope)
			return true;
		if(Scope parent = sub.parent())
			return subscope(scope, parent);
	}
	return false;
}


void
Decl::add()
{
	Scope sc = scope();
	Id cid = id();
	if(!cid)
		cid = ID_ANON;
	_link = 0;
	if(!sc.add(cid, this)) {
		Decl *d = sc.get(cid);
		_link = d;
		sc.set(cid, this);
	}
}

void
Decl::rename(Id newid)
{
	printf("rename\n");
	Id oldid = id();
	if(!oldid)
		oldid = ID_ANON;
	Decl *d = scope().get(oldid);
	if(d == this) {
		scope().del(oldid);
		if(_link)
			scope().add(oldid, _link);
	} else {
		while(d->_link != this)
			d = d->_link;
		d->_link = _link;
	}
	_id = newid;
	add();
	printf("rename done\n");
}


Decl *
Decl::parent() const
{
	Decl *pd = 0;
	if(dtype() == TDecl::dt_enumval) {
		if(Enum *e = dcast_Enum(rtype())) {
			pd = e->decl();
		} /* else
			INTERNAL_ERROR; */
	} else if((pd = scope().decl())) {
		/*
		if(scope() != global()
		   && (!pd->ptype() || !pd->ptype()->istype(Type_Aggr)))
			pd = 0;
		*/
	}
	return pd;
}

void
Decl::chain()
{
	if(Decl *pd = parent()) {
		Decl **dpp = &pd->_a.first;
		while(*dpp)
			dpp = &(*dpp)->_next;
		(*dpp) = this;
		if(field())
			pd->_a.tsize++;
	}
}

void
Decl::unchain()
{
	if(Decl *pd = parent()) {
		Decl **dpp = &pd->_a.first;
		while(*dpp && *dpp != this)
			dpp = &(*dpp)->_next;
		// assert(*dpp == d);
		*dpp = _next;
		if(field()) 
			pd->_a.tsize--;
		_next = 0;
	}
}


Decl*
Decl::clone(Scope sc, Id id) const
{
	Decl *dn = new Decl(*this);
	if(id)
		dn->_id = id;
	dn->where(-1);
	dn->_link = 0;
	dn->_next = 0;
	dn->_scope = sc;
	dn->add();
	dn->chain();
	return dn;
}

bool
Decl::implements(Decl *dt) const
{
	// assert(dt->isaggr());
	if(this == dt)
		return true;
	if(!afirst())
		return false;
	// XXX - possible infinite loop problem?
	for(Decl *dm = afirst(); dm; dm = dm->next()) {
		if(dm->dtype() == Decl::dt_impl) {
			if(!dt)
				return true;
			if(dt->idecl()->implements(dt))
				return true;
		}	
	}
	// Scope sc = getaggrscope(this);
	Scope sc = afirst()->scope();
	for(Decl *di = sc.get(ID_INHERIT); di; di = di->link())
		if(di->idecl()->implements(dt))
			return true;
	return false;
}


bool
Decl::istype() const
{
	if(storage() == Decl::sc_typedef)
		return true;
	if(dtype() == Decl::dt_enum) 
		return true;
	if(dtscope())
		return true;
	return false; 
}

bool
Decl::declarable() const
{
	if(dtype() == Decl::dt_enum) 
		return false;
	if(storage ()== Decl::sc_typedef) 
		return tagged();
	if(isfunc())
		return true;
	if(external())
		return true;
	if(scope().parent()) 
		return storage() == Decl::sc_static;
	else 
		return storage() == Decl::sc_default;
}

