/*   $Id: inttypes.h,v 1.2 1999/07/16 20:18:50 shaggy Exp $
 *	integer type definitions for big-endian architectures
 *      relative path: ./lib/m-pc
 */ 

/* Note:
 *	I'll probably get a lot of nasty letters for this, but I've made
 *	a minor change to the "standard" I adopted for this file.
 *	All of the basic type declarations there had "_t" appended to
 *	them, such as "uint16_t".  Although I follow this convention
 *	for most of my typedef'ed types (such as "vaddr_t"), I consider
 *	these bittypes so basic to all definitions that the "_t" is unnecessary,
 *	ugly, and inconvenient.   Most people wouldn't agree with writing 
 *	'int_t' and 'double_t' everywhere, so why clutter up everything
 *	for other types whose use is just as basic and common? 
 *
 *	With enough criticism and good reasons, I may be able to be
 *	convinced to change my point of view on this, but then again,
 *	with enough support, I may be convinced to stand my ground.
 *
 *	I really would like to see the integer problem standardized,
 *	and when it is formally adopted, I will use whatever standard 
 *	comes out, but until that time, I would like to push for the
 *	simpler, shorter, cleaner syntax that does not have the '_t'
 *	for the basic integer types, and reserves that notation for 
 *	other specialized types that are more likely to be confused.
 *	(such as 'ptrdiff_t', etc)
 *
 *	(Of course, what I would REALLY like to see is an extension to the
 *	C and C++ languages that has built in support for specifying the 
 *	size, but that will take quite some time to become a standard)	
 */


#ifndef ASM

#include <btypes.h>

typedef _int8	int8;
typedef _int16	int16;
typedef _int32	int32;
typedef _int64	int64;

typedef _uint8	uint8;
typedef _uint16	uint16;
typedef _uint32	uint32;
typedef _uint64	uint64;

typedef _int64	intmax;		/* largest signed integer */
typedef _uint64	uintmax;	/* largest unsigned integer */

typedef _int32	intptr;		/* signed integer that can hold a void*  */
typedef _uint32	uintptr;	/* unsigned integer that can hold a void*  */

typedef int		intfast;	/* most efficient signed integer */
typedef unsigned int	uintfast;	/* most efficient unsigned integer */


typedef _int8	int_least8;
typedef int	int_fast8;
typedef	_int16	int_least16;
typedef int	int_fast16;
typedef _int32	int_least32;
typedef int	int_fast32;
typedef _int64	int_least64;
typedef _int64	int_fast64;

typedef _uint8		uint_least8;
typedef unsigned int	uint_fast8;
typedef	_uint16		uint_least16;
typedef unsigned int	uint_fast16;
typedef _uint32		uint_least32;
typedef unsigned int	uint_fast32;
typedef _uint64		uint_least64;
typedef _uint64		uint_fast64;

#define INT8_MIN		(-128)
#define INT16_MIN		(-32767-1)
#define INT32_MIN		(-2147483647-1)
#define INT64_MIN		(-9223372036854775807-1)

#define INT_LEAST8_MIN		INT8_MIN
#define INT_LEAST16_MIN		INT16_MIN
#define INT_LEAST32_MIN		INT32_MIN
#define INT_LEAST64_MIN		INT64_MIN

#define INT_FAST8_MIN		INT32_MIN
#define INT_FAST16_MIN		INT32_MIN
#define INT_FAST32_MIN		INT32_MIN
#define INT_FAST64_MIN		INT64_MIN

#define INT8_MAX		(127)
#define INT16_MAX		(32767)
#define INT32_MAX		(2147483647)
#define INT64_MAX		(9223372036854775807)

#define UINT8_MAX		(255U)
#define UINT16_MAX		(65535U)
#define UINT32_MAX		(4294967295U)
#define UINT64_MAX		(18446744073709551615U)

#define INTMAX_MIN		INT64_MIN
#define INTMAX_MAX		INT64_MAX
#define UINTMAX_MAX		UINT64_MAX

#define INTFAST_MIN		INT32_MIN
#define INTFAST_MAX		INT32_MAX
#define UINTFAST_MAX		UINT32_MAX

#define INT_LEAST8_MAX		INT8_MAX
#define INT_LEAST16_MAX		INT16_MAX
#define INT_LEAST32_MAX		INT32_MAX
#define INT_LEAST64_MAX		INT64_MAX

#define INT_FAST8_MAX		INT32_MAX
#define INT_FAST16_MAX		INT32_MAX
#define INT_FAST32_MAX		INT32_MAX
#define INT_FAST64_MAX		INT64_MAX

#define UINT_LEAST8_MAX		UINT8_MAX
#define UINT_LEAST16_MAX	UINT16_MAX
#define UINT_LEAST32_MAX	UINT32_MAX
#define UINT_LEAST64_MAX	UINT64_MAX

#define UINT_FAST8_MAX		UINT32_MAX
#define UINT_FAST16_MAX		UINT32_MAX
#define UINT_FAST32_MAX		UINT32_MAX
#define UINT_FAST64_MAX		UINT64_MAX


#define INTPTR_MIN		INT32_MIN
#define INTPTR_MAX		INT32_MAX
#define UINTPTR_MAX		UINT32_MAX


#define __CONCAT__(a,b)		a##b

#define INT8_C(c)		(c)
#define INT16_C(c)		(c)
#define INT32_C(c)		(c)
#define INT64_C(c)		__CONCAT__(c,ll)

#define UINT8_C(c)		__CONCAT__(c,u)
#define UINT16_C(c)		__CONCAT__(c,u)
#define UINT32_C(c)		__CONCAT__(c,u)
#define UINT64_C(c)		__CONCAT__(c,ull)

#define INTMAX_C(c)		__CONCAT__(c,ll)
#define UINTMAX_C(c)		__CONCAT__(c,ull)


#endif

