/*  $Id: deco.cc,v 1.167 2000/01/04 05:09:35 shaggy Exp $
 *      Dynamic Encapsulator of C++ Objects
 *      
 *      relative path: util/deco
 *  
 *  Copyright (c) 1996-1998 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define deco_VER_MAJOR	0
#define deco_VER_MINOR	13
#define deco_VER_PATCH	0
#define deco_VER_STR	"0.13"
#define deco_VER_DATE	"January 03, 2000"
#define deco_VER_NDATE	"01/03/00"

// #pragma Make Config_Needed += local
#pragma Make Includes_Banned += -nostdinc -nostdinc++
#pragma Make Flags_Banned += -static

#define DEBUG
#define CACTION_DEBUG

#include <libc_posix.h>
#include <stdx1.h>
#include <assert.h>
#include <defines>
#include <ldefines>

#include <CObj>
#include <CLex>
#include <StrTab>
#include <SymTab>
#include <ParsePath>
#include <vCAction>
#include <CParse>
#include <CRep>
#include <COut>
#include <CAction>

// XXX - temporary fix:
extern "C" FILE *popen(const char *cmd, const char *type);
extern "C" int pclose(FILE *f);


bool debugaction = false;
bool printexec = false;

#define INTERNAL_ERROR  { internal_error(__FILE__, __LINE__, __FUNCTION__); \
			  for(;;); }

#include "decoconf.h"


// XXX - ideas:
//	- extend Item structure by making it a class that can
//	  be derived from and have methods added to
//	- extend Token by putting base definition in a file that can
//	  be included and then have more entries added to
//	- make sure specific/generic print routines are available for it 
//	  too

// XXX - at some point, any knowledge specific to certain data
// types should be removed from DECO, since it is supposed to be 
// a generic C++ pre-processor that generates code that can compile
// on any architecture. 


// Need to clean up use of Item by providing a local
// Item class that has a separate structure for each token ID,
// so that casting won't be necessary and the expected layout for
// each token is clear and can be checked for errors


// lexical shortcuts:
typedef vStrTab::Id Id;


Id _id_instance, _id_dtobj, _id_dcobj, _id_dataclass;
Id _id_dbind, _id_dcbind, _id_dcheck;
Id _id_o, _id_btrp;
Id _id_D_Type, _id_D_Class;
Id _id__objp;
Id _id__itabp;
Id _id__itable;
Id _id_this_ref;
Id _id_this_r;
Id _id_this_w;
Id _id_this_rw;
Id _id_btr;
Id _id_copy;
Id _id_share;
Id _id__iobj, _id_iobj;
Id _id__cobj, _id_cobj;
Id _id_any;
Id _id__type;

// declarations from dcobj.hh:
Decl *_d_BTRef;
Decl *_d_bt__DCBind;
Decl *_d_bt__this_r;
Decl *_d_bt__this_w;
Decl *_d_bt__this_rw;
Decl *_d_bt__this_ref;
Decl *_d_bt__iobj;
Decl *_d_bt__cobj;
Decl *_d_any;
Decl *_d_Obj;
Decl *_d_obj__init_alloc;
Decl *_d_obj__init_ref;
Decl *_d_obj__this_r;
Decl *_d_obj__this_w;
Decl *_d_obj__this_rw;
Decl *_d_obj__this_ref;

Decl *_d_D_Class;
Decl *_d_D_Class_dcobj;
Decl *_d_D_Type;
Decl *_d_D_Type_dcobj;


struct Itable {
	vSymTab *symtab;
	Decl *ditab;
	ExprList *eitab;
	int last;
	Scope sc;
};


class Deco : public CAction {
private:
	int _input_fd;
	char _buf[4096];


	bool _dcxx_initialized;
	bool _dcxxobj_initialized;

	Decl* declfind(csz *name, Scope sc = 0);
	void dcxx_init();
	bool dcxxobj_init();

public:
	Deco(CLex &clp, vStrTab &strtab, vSymTab &symtab);

	bool open_input(csz *fname = 0, csz *cpp = 0);
	void close_input();
	void get_input(bool split_token);
	void abort_exit();
	void internal_error_exit();
	void veprintf(csz *fmt, va_list args);

	void dcxx_to_cxx();
	void dcxx_to_cxx(Scope, bool fn);
	void type_bases(Decl *dt, Scope sc, int *b);
	Decl* cxx_of_type(Decl *);
	Decl* cxx_of_dtype(Decl *);
	Decl* cxx_of_class(Decl *);
	Decl* cxx_of_dclass(Decl *);
	Decl* cxx_of_decl(Decl *, bool fn);
	Decl* impl_to_cxx(Decl *ddi, Decl *dc);
	void typeimpl_to_cxx(Decl *ddi, Decl *dc);
	void dclassimpl_to_cxx(Decl *ddi, Decl *dc, Decl *ddat);
	void type_mfn_to_cxx(Decl *, Scope, int b, int n);
	type_t cxx_to_c_fn(type_t cxxf);
	int addimpl(class Itable *itab, Decl *ddi, Decl *dti);
	Decl* gen_itable(Decl *ddi, Scope sci);
	void gen_dbind(Decl *dc, Itable *itab, Decl *dti);
	void gen_dcbind(Decl *dc, Decl *dci, Decl *ditab, Decl *ddcheck,
								type_t ti);
	void gen_binder(Decl *dc, Decl *dci, Itable *itab, Decl *dti);
	Decl* gen_conv_op(type_t, Scope sc, Decl *ddbind);
	Decl* gen_conv_relay(Decl *dm, Decl *db, int b, Scope sc);
	Decl* gen_thunk(Decl *dm, Scope, Id id = 0);
	Decl* newmethod(Id id, Scope sc, type_t rt, bool qconst = false,
						int nargs = 0, Param * = 0);
	Expr expr_mfnadr(type_t newtype, Decl *d);
	Expr expr_arradr(type_t newtype, Decl *d, int i = 0);
	Stmt stmt_relaycall(Expr e, const FnInfo *fi);
	Stmt stmt_this_relaycall(Decl *th, Expr e, const FnInfo *fi);
	ExprList copyargs(ExprList el, Func *f);
	void localize_params(Decl *d);

	void predefine();

	Id id_encode(csz *prefix, csz *name);

	action EndAggr;
};



Deco::Deco(CLex &clp, vStrTab &strtab, vSymTab &symtab)
	 : CAction(clp, strtab, symtab)
{
	_dcxx_initialized = false;
	_dcxxobj_initialized = false;
}

void
Deco::localize_params(Decl *d)
{
	FnInfo fi;
	assert(getfninfo(d->ptype(), &fi));
	Scope pscope = newscope(d->scope(), 0);
	Param p;
	for(uint i=0; i<fi.f->nparam(); i++) {
		Decl *pd = param_decl(fi.f, i);
		p.type = fi.f->ptype(i);
		p.name = pd->id();
		p.defval.token = NONE;
		addparam(i, pscope, p);
	}
	if(fi.m) 
		d->ptype(type_method(d->scope().decl()->rtype(),
				type_func(fi.r, fi.p, pscope)));
	else
		d->ptype(type_func(fi.r, fi.p, pscope));
	d->rtype(realtype(d->ptype()));
}


void
Deco::EndAggr(Item *r, Item_ agsc, Item_ scope, Item_)
{
    	PRINT_ACTION(r, agsc, scope, none);
	CAction::EndAggr(r, agsc, scope, none);
	Decl *d = uAggrId(r).decl;
	Scope sd = getaggrscope(d);
	bool copyctors = true;
	Decl *dtype = 0;
	Decl *dbi = sd.get(ID_INHERIT);
	if(!dbi || dbi->link())
		copyctors = false;
	for(Decl *di = d->afirst(); di; di = di->next()) {
		if(di->field() || di->isctor()) {
			copyctors = false;
		} else if(di->dtype() == Decl::dt_impl) {
			for(Decl *dm = di->afirst(); dm; dm = dm->next()) {
				FnInfo fi;
				if(!getfninfo(dm->rtype(), &fi)) {
					error("non-function `%s' declared "
							"in implementation of type `%s'",
							cstr(dm->id()), cstr(di->id()));
					continue;
				}
				Decl *dtm = getaggrscope(di->idecl()).ufind(dm->id());
				if(!dtm) {
					error("function `%s' not part of type `%s'",
							cstr(dm->id()), cstr(di->id()));
					continue;
				}
				if(!findfn(dtm, &fi)) {
					sz b1[MAXTYPELEN];
					sz b2[MAXIDLEN];
					strcpy(b1, cstr(dm->id()));
					scoped_string(dtm->scope().decl(), b2);
					error("parameters of `%s' don't match any in `%s'",
							type_string(dm->rtype(), b1), b2);
					continue;
				}
			}
			if(!dtype) {
				dtype = forwref(Decl::dt_type, _id__type, sd);
				newaggrscope(dtype);
				dtype->defined(true);
			}
			inherit(dtype, di->idecl(), TDecl::ac_public);
		} 
	}
	if(copyctors) {
		Decl *db = dbi->idecl();
		for(Decl *dm = db->afirst(); dm; dm = dm->next()) {
			if(dm->isctor()) {
				FnInfo fi;
				getfninfo(dm->rtype(), &fi);
				Decl *dn = dm->clone(sd, d->id());
				StmtList slctor(fi.sc);
				slctor += stmt_relaycall(Expr(db), &fi);
				dn->vexpr(ExprCtor(slctor, StmtList())).sinline(true);
				dn->otype(Decl::OT_IMPLICIT);
			}
		}
	}

	if(d->dtype() == Decl::dt_type || d->dtype() == Decl::dt_dtype) {
		if(!_d_any) {
			_d_any = declfind("any");
			if(!_d_any) 
				abort("`any' must be first type declared");
		}
		if(!dbi && d != _d_any)
			inherit(d, _d_any, TDecl::ac_public);
	}

	if(d->dtype() != Decl::dt_dclass)
		return;

	// For classes:
	// create 'dataclass', inheriting from bases (count bases)
	// map all functions and fields into dataclass
	// loop through all implementation scopes
	//	map each function to dataclass as type__fn,

}


/////////////////////////////////////////////////////////////////////////////
// code re-writing

class ExprRCAST : public Expr { public: ExprRCAST(type_t t, Expr e)
			: Expr(T(KW_RCAST), t, e) { rtype(realtype(t)); }};
class ExprSCAST : public Expr { public: ExprSCAST(type_t t, Expr e)
			: Expr(T(KW_SCAST), t, e) { rtype(realtype(t)); }};
class ExprCCAST : public Expr { public: ExprCCAST(type_t t, Expr e)
			: Expr(T(KW_CCAST), t, e) { rtype(realtype(t)); }};
class ExprCAST : public Expr { public: ExprCAST(type_t t, Expr e)
			: Expr(T(OP_CAST), t, e) { rtype(realtype(t)); }};


static Expr
expr_rewrite(Expr e, Expr (*rwf)(Expr, Expr), Expr p)
{
	Expr ne;
	if(!e)
		return e;
	// printf("eee=%p, %d, (%p, %p, %p) %p\n", (ExprNode*)e, e.op(), (ExprNode*)e.left(), (ExprNode*)e.mid(), (ExprNode*)e.right(), (ExprNode*)e.next());
	switch(e.op()) {
	case T(VAL_ID):
	case T(TOKENLIST):
	case T(LITERAL):
	case T(ID):
	case T(TYPE):
		ne = e;
		break;
	case T(EXPRLIST):
	case T(STMTLIST): {
		Expr eleft = expr_rewrite(e.left(), rwf, p);
		if(e.left().is(eleft))
			ne = e;
		else {
			Expr i = eleft;
			while(i.next())
				i = i.next();
			ne = Expr(e.op(), eleft, e.mid(), i);
		}
		} break;
	default: {
		Expr eleft = expr_rewrite(e.left(), rwf, p);
		Expr emid = expr_rewrite(e.mid(), rwf, p);
		Expr eright = expr_rewrite(e.right(), rwf, p);
		if(e.left().is(eleft) && e.mid().is(emid)
				&& e.right().is(eright))
			ne = e;
		else {
			ne = Expr(e.op(), eleft, emid, eright);
			ne.rtype(e.rtype());
		}}
	}
	Expr erw = rwf(ne, p);
	Expr enext = expr_rewrite(e.next(), rwf, p);
	if(!erw.next().is(enext)) {
		if(ne.is(e) || !ne.is(erw))
			ne = erw.dup();
		ne += enext;
	} else
		ne = erw;
	return ne;
}



/////////////////////////////////////////////////////////////////////////////
// DC++ to C++ conversions

// a lot of the following functions are rather grotesque
// hopefully, they will get trimmed down once better support functionality
// exists, since there's a lot of redundant code here
// The code here also employs "brute force" far too often.
// Something more elegant, but functionally equivalent, needs to replace it.
//
// Believe it or not, it's actually getting better already... 
// You should have seen it before some of this stuff got split
// into separate functions...

/*
static ExprList
elist()
{
	return ExprList();
}
*/

static ExprList
elist(Expr e1)
{
	return ExprList() += e1;
}

/*
static ExprList
elist(Expr e1, Expr e2)
{
	return (ExprList() += e1) += e2;
}
*/

/*
static ExprList
elist(Expr e1, Expr e2, Expr e3)
{
	return ((ExprList() += e1) += e2) += e3;
}
*/

ExprList
Deco::copyargs(ExprList p, Func *f)
{
	for(uint i=0; i<f->nparam(); i++) {
		Decl *pd = param_decl(f, i);
		if(!pd || !pd->id()) {
			sz buf[10];
			sprintf(buf, "_a%d_", i);
			Id id = this->id(buf);
			if(pd)
				pd->rename(id);
			else {
				Param np;
				np.type = f->ptype(i);
				np.name = id;
				np.defval.token = NONE;
				pd = addparam(i, f->scope(), np);
			}
		}
		p += Expr(pd);
	}
	return p;
}

/*
static Expr
rewrite_id_expr(Expr e, Expr p)
{
	if(e.op() == T(VAL_ID) && e.decl() == p.decl()) 
		return p.next();
	else
		return e;
}
*/


static Expr
rewrite_dcxx_to_cxx(Expr e, Expr)
{	
	e.rtype(convtype(e.rtype()));
	if(e.op() == T(OP_CONDASN)) {
		return Expr(T(OP_ASN), e.left(), e.right());
	} else if(e.op() == T(VAL_ID)) {
		Decl *d = e.decl();
		if(d->conv()) {
			while(d->conv())
				d = d->conv();
			return Expr(d);
		}
		d->ptype(convtype(d->ptype()));
		d->rtype(convtype(d->rtype()));
		e.rtype(d->rtype());
	} else if(e.op() == T(DECL)) {
		Decl *d = e.left().decl();
		if(d->condasn()) {
			d->condasn(false);
			type_t t = convtype(Expr(d->vexpr()).rtype());
			Expr(d->vexpr()).rtype(t);
			Aggr *aggr = dcast_Aggr(t);
			Decl *ddcbind = Scope(aggr->scope()).ufind(_id_dcbind);
			Expr ef = Expr(T(OP_DOT), d->vexpr(), ddcbind).
				rtype(convtype(ddcbind->rtype()));

			// XXX: check original
			d->vexpr(ExprCall(ef, elist(d))).rtype(convtype(d->rtype()));
		}
	} else if(Aggr *aggr = dcast_Aggr(e.rtype())) {
		Scope sc = Scope(aggr->scope());
		if(sc) {
			Decl *dc = sc.decl();
			if(dc->conv()) 
				return Expr(e).rtype(dc->next()->rtype());
		}
	}
	return e;	
}


// some little naming conventions, to try to keep things
// from getting too confusing:
// (yes, this borders on hungarian notation in some ways,
// which I personally can't stand, but given the size and
// complexity of some of these functions, I figured this
// would help avoid a bunch of errors-- of course, the better
// solution would be to break this up into enough smaller
// pieces so that one and two letter names would be sufficient)
//
//	d_XXX		Decl* for XXX
//	t_XXX		type_t for XXX
//	s_XXX		scope owned by XXX (*not* parent scope)
//	ps_XXX		parent scope for XXX
//	id_XXX		vStrTab::Id of XXX
//	f_XXX		Func* for XXX
//	m_XXX		Method* for XXX
//	e_XXX		Expr
//	el_XXX		Expression List
//	sl_XXX		Statement List
//	XXX_c		refers to class being constucted from type/dclass
//	XXX_dt		refers to dynamic type being converted
//	XXX_dc		refers to dynamic class being converted
//	XXX_dat		refers to dataclass (inside of dclass)
//	XXX_m		refers to current method
//	XXX_di		refers to dynamic implementation subclass
//	XXX_ci		refers to implementation subclass (converted)
//	XXX_dti		refers to dynamic type being implemented
//	XXX_cti		refers to class of type being implemented (converted)
//	XXX_itab	refers to itable[]
//	XXX_YYY		refers to YYY
//
// The usual convention applies about prefixing an identifier with _
// to show that it's a field.  Also, for local vars with limited
// scope, it's reasonable to drop the _ in the middle.

Decl*
Deco::newmethod(Id id, Scope s, type_t rt, bool qconst = false, int np, Param *param)
{
	type_t ptype[8];
	Scope pscope = newscope(s, 0);
	for(int i=0; i<np; i++) {
		ptype[i] = param[i].type;
		addparam(i, pscope, param[i]);
	}
	type_t psig = type_psig(np, ptype);
	type_t ot = s.decl()->rtype();
	if(qconst)
		ot = type_const(ot);
	Decl *d = newdecl(id, s, type_method(ot, type_func(rt, psig, pscope)));
	addthis(d);
	d->fpscope(pscope);
	assert(d->isfunc());
	assert(d->ismeth());
	assert(!d->field());
	return d;
}


// return an expression that gets the address of a method function
// and converts it to the specified type
Expr
Deco::expr_mfnadr(type_t newtype, Decl *d)
{
	Expr emptr = Expr(OP_REF, Expr(d));

	Decl *dl = d->scope().uget(d->id())->link();
	if(dl && !dl->tagged())
		emptr = ExprSCAST(type_ptrto(d->rtype()), emptr);
	return ExprCAST(newtype, emptr);
}

// return an expression that gets the address of an index into an array
// and converts it to the specified type
Expr
Deco::expr_arradr(type_t newtype, Decl *d, int i)
{
	Expr e;
	if(i == 0) 
		e = Expr(d);
	else
		e = Expr(OP_REF, Expr(OP_INDEX, Expr(d), Lit(&Type_int, i)));
	if(newtype)
		return ExprRCAST(newtype, e);
	else
		return e;
}

// return a statement that relays a call to the given expression,
// using all of the passed arguments of the specified function
Stmt
Deco::stmt_relaycall(Expr e, const FnInfo *fi)
{
	ExprList elp = copyargs(ExprList(), fi->f);
	Expr ec = ExprCall(e, elp);
	if(fi->r && fi->r != &Type_void)
		return StmtReturn(ec);
	else
		return Stmt(ec);
}

// return a statement that relays a call to the given expression,
// using all of the passed arguments of the specified function,
// preceeded by an explicit 'this'
Stmt
Deco::stmt_this_relaycall(Decl *th, Expr e, const FnInfo *fi)
{
	ExprList elp = copyargs(ExprList() += Expr(th), fi->f);
	Expr ec = ExprCall(e, elp);
	if(fi->r && fi->r != &Type_void)
		return StmtReturn(ec);
	else
		return Stmt(ec);
}

void
Deco::gen_dcbind(Decl *dc, Decl *dci, Decl *ditab, Decl *ddcheck, type_t ti)
{
	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;
	Scope sc = getaggrscope(dc);

	// gen: bool c::dcbind(::ti*) const
	// {
	//	btrp->_itabp = ti::_itable;
	//	btrp->_objp = const_cast<c*>(this);
	//	if(static_cast<ti*>(btrp)->dcheck())
	//		return true;
	//	else {
	//		btrp->_itabp = 0;
	//		btrp->_objp = 0;
	//		return false;
	//	};
	param[0].name = _id_btrp;
	param[0].type = type_ptrto(ti);
	Decl *ddcbind = newmethod(_id_dcbind, sc, &Type_bool, true, 1, param);
	StmtList sl(ddcbind->fpscope());
	Decl *dbtrp = Scope(ddcbind->fpscope()).get(_id_btrp);
	Decl *dobjp = getaggrscope(_d_BTRef).ufind(_id__objp);
	Decl *ditabp = getaggrscope(_d_BTRef).ufind(_id__itabp);
	assert(dbtrp);
	assert(dobjp);
	assert(ditabp);
	sl += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(ditabp)),
		Expr(ditab)));
	sl += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(dobjp)),
		ExprRCAST(type_ptrto(type_declref(_d_Obj)),
			ExprCCAST(type_ptrto(type_declref(dc)),
				getthis(ddcbind)))));
	Expr ec = ExprCall(Expr(OP_PTRTO,
			ExprRCAST(type_ptrto(type_declref(dci)), dbtrp),
						ddcheck),
			ExprList());
	StmtList slelse(ddcbind->fpscope());
	sl += StmtIfElse(ec, StmtReturn(Lit(&Type_bool, 1)), Stmt(slelse));
	slelse += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(ditabp)),
		Lit(&Type_int, 0)));
	slelse += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(dobjp)),
		Lit(&Type_int, 0)));
	slelse += StmtReturn(Lit(&Type_bool, 0));
	ddcbind->vexpr(sl);

	// gen: ti c::dcbind(::ti&) const return btr; { dcbind(&btr); } 
	// gen: ti c::dcbind(::ti&) const {::ti btr; dcbind(&btr); return btr;}
	param[0].name = 0;
	param[0].type = type_refto(ti);
	Decl *ddcbind2 = newmethod(_id_dcbind, sc, ti, true, 1, param);
	StmtList sl2(ddcbind2->fpscope());
	Decl *d = newdecl(_id_btr, ddcbind2->fpscope(), ti);
	sl2 += Stmt(ExprDecl(d));
	sl2 += StmtCall(ddcbind, elist(Expr(OP_REF, d)));
	sl2 += StmtReturn(d);
	ddcbind2->vexpr(sl2).sinline(true);
}

void
Deco::gen_dbind(Decl */*dc*/, Itable *itab, Decl *dti)
{
	type_t ti = type_declref(dti->conv());

	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;

	// gen: void dbind(ti *btrp) const 
	//	{ btrp->_iobj = this;
	// 	  btrp->_itabp = &_itable[n]; }
	param[0].name = _id_btrp;
	param[0].type = type_ptrto(ti);
	Decl *ddbind = newmethod(_id_dbind, itab->sc, &Type_void, true, 1, param);
	StmtList sl3(ddbind->fpscope());
	Decl *dbtrp = Scope(ddbind->fpscope()).get(_id_btrp);
	Decl *diobj = getaggrscope(_d_BTRef).ufind(_id__iobj);
	Decl *ditabp = getaggrscope(_d_BTRef).ufind(_id__itabp);
	assert(dbtrp);
	assert(diobj);
	assert(ditabp);
	sl3 += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(diobj)),
			ExprRCAST(type_ptrto(&Type_void), getthis(ddbind))));
	sl3 += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(ditabp)),
		expr_arradr(0, itab->ditab, itab->last)));
	ddbind->vexpr(sl3);

	gen_conv_op(ti, itab->sc, ddbind);
}


void
Deco::gen_binder(Decl *dc, Decl */*dci*/, Itable *itab, Decl *dti)
{
	// Scope sci = getaggrscope(dci);
	// Decl *ddcheck = sci.uget(_id_dcheck);
	//if(ddcheck)
	//	gen_dcbind(dc, dci, ditab, ddcheck, ti);
	// else
		gen_dbind(dc, itab, dti);
}

Decl*
Deco::impl_to_cxx(Decl *ddi, Decl *dc)
{
	if(!ddi->idecl()->defined())
		return 0;
	Decl *dci = forwref(Decl::dt_class, ddi->id(), getaggrscope(dc));
	Scope sci = newaggrscope(dci);

	// gen: : public dti;
	assert(ddi->idecl()->conv());
	Decl *dtic = ddi->idecl()->conv();
	inherit(dci, dtic, TDecl::ac_public);

	// gen: private: inline ci() {}
	newmethod(ddi->id(), sci, 0)->vexpr(StmtList())
				.access(Decl::ac_private).sinline(true);

	// gen: private: friend class c;
	newdecl(ID_FRIEND, sci, type_declref(dc))->storage(Decl::sc_friend);

	// gen: c* iobj() { return reinterpret_cast<c*>(_iobj); }
	type_t t = type_ptrto(type_declref(dc));
	StmtList sl;
	sl += StmtReturn(ExprRCAST(t, _d_bt__iobj));
	newmethod(_id_iobj, sci, t)->vexpr(sl).sinline(true);

	return dci;
}


Id
Deco::id_encode(csz *prefix, csz *name)
{
	sz buf[MAXIDLEN];
	strcpy(buf, prefix);
	char *bp = buf+strlen(buf);
	while(char c = *name) {
		if((c >= 'a' && c <= 'z') 
		  || (c >= 'A' && c <= 'Z')
		  || (c >= '0' && c <= '9')
		  || (c == '_'))
			*bp++ = c;
		else
			bp += sprintf(bp, "_0x%02X_", c);
		++name;
	}
	*bp = '\0';
	return id(buf);
}

Decl*
Deco::gen_conv_relay(Decl *dm, Decl *dd, int b, Scope sc)
{
	Decl *db = dm->scope().decl()->conv();
	assert(db);
	
	printf("conv relay\n");

	// gen: { ti btr = *this; btr._itabp += #;
	// 	return btr->default__xxx(args); }

	FnInfo fi;
	assert(getfninfo(dd->rtype(), &fi));
	// dd->initialized(true);
	Scope scf = newscope(fi.sc, 0);
	StmtList sl(scf);
	Decl *dthis = getthis(dd);
	type_t ti = type_declref(db);
	printf("sc=%p ti=%p\n", (vSymTab*)dd->fpscope(), ti);
	Decl *d = newdecl(_id_btr, scf, ti);
	d->vexpr(Expr(OP_PTR, dthis));
	sl += Stmt(ExprDecl(d));
	sl += Stmt(Expr(OP_ASNADD, Expr(OP_DOT, d, sc.uget(_id__itabp)),
			Lit(&Type_int, b)));
	Expr e = Expr(OP_DOT, d, dm);
	sl += stmt_relaycall(e, &fi);
	dd->vexpr(sl);
	return dd;
}

Decl*
Deco::gen_thunk(Decl *dm, Scope sc, Id id)
{
	Decl *dn = dm->clone(sc, id);
	localize_params(dn);
	FnInfo fi;
	assert(getfninfo(dn->rtype(), &fi));
	Expr e;
	e = ExprSCAST(dm->scope().decl()->rtype(), Expr(OP_PTR, getthis(dn)));
	e = Expr(OP_DOT, e, dm);
	dn->vexpr(StmtList(fi.sc) += stmt_relaycall(e, &fi)).sinline(true);
	return dn;
}

#if 0
	if(b >= 0) {
		/*
		Decl *d = dm->scope.ufind(id_def);
		assert(d);
		d = findfn(d, dm->rtype);
		assert(d);
		gen_conv_relay(d, dd, b, sc);
		*/
	}
#endif

type_t
Deco::cxx_to_c_fn(type_t cxxf)
{
	Method *meth = dcast_Method(cxxf);
	assert(meth);
	Func *func = dcast_Func(meth->ftype());
	assert(func);
	Params *par = dcast_Params(func->psig());
	assert(par);
	type_t p[par->nparam()+1];
	for(uint i=0; i<par->nparam(); i++)
		p[i+1] = par->ptype(i);
	p[0] = type_ptrto(meth->otype());
	type_t psig = type_psig(par->nparam()+1, p, par->variable());
	return type_func(func->rtype(), psig, Scope(func->scope()->parent()));
}

void
Deco::type_mfn_to_cxx(Decl *dm, Scope sc, int b, int n)
{
	Decl *d = sc.uget(dm->id());
	d = findfn(d, dm->rtype());
	if(d) {
		warning("nuking duplicate");
		d->dtype(Decl::dt_default);
		d->ptype(0);
		d->rtype(0);
		d->unchain();
		return;
	}
	d = dm->clone(sc);
	dm->conv(d);

	Method *m = dcast_Method(d->ptype());
	assert(m);
	d->ptype(type_method(type_declref(sc.decl()), m->ftype()));
	d->rtype(realtype(d->ptype()));
	FnInfo fi;
	assert(getfninfo(d->rtype(), &fi));
	// d->initialized(true);
	StmtList sl(fi.sc);
	Decl *ditab = sc.uget(_id__itabp);
	Expr eitab = ditab;
	if(b >= 0)
		eitab = ExprRCAST(ditab->rtype(),
			Expr(OP_INDEX, eitab, Lit(&Type_int, b)));
	Expr e = Expr(OP_INDEX, eitab, Lit(&Type_int, n));
	e = ExprCAST(type_ptrto(cxx_to_c_fn(d->ptype())), e);
	// e = Expr(OP_MPTRTO, Expr(getthis(d)), e);
	sl += stmt_this_relaycall(getthis(d), e, &fi);
	d->vexpr(sl).sinline(true);
	d->where(n);
}



int
Deco::addimpl(Itable *itab, Decl *ddi, Decl *dti)
{
	Decl *dcti = cxx_of_type(dti);
	Id id = Id(dcti);
	vSymTab::Val val;
	int idx;
	if(itab->symtab->get(id, val))
		return int(val);
	Scope sdi = getaggrscope(ddi);
	Scope sti = getaggrscope(dti);
	Scope scti = getaggrscope(dcti);

	Decl *dinh = sti.get(ID_INHERIT);
	for(Decl *dbi = dinh; dbi; dbi = dbi->link()) 
		 addimpl(itab, ddi, dbi->idecl());

	idx = itab->last;
	val = vSymTab::Val(idx);
	itab->symtab->add(id, val);

	gen_binder(0, 0, itab, dti);

	for(Decl *dm = dti->afirst(); dm; dm = dm->next()) {
		assert(dm->ismeth());
		Decl *d = sdi.ufind(dm->id());
		if(!(d = findfn(d, dm->rtype()))) {
			d = scti.ufind(id_encode("default__", cstr(dm->id())));
			d = findfn(d, dm->rtype());
			d = gen_thunk(d, itab->sc);
		}
		assert(d);
		// XXX - need to match params
		while(d->conv())
			d = d->conv();
		*(itab->eitab) += expr_mfnadr(type_ptrto(&Type_void), d);
		++(itab->last);
	}
	for(Decl *dbi = dinh; dbi; dbi = dbi->link()) {
		int i = addimpl(itab, ddi, dbi->idecl());
		*(itab->eitab) += expr_arradr(type_ptrto(&Type_void),
					 itab->ditab, i);
		++(itab->last);
	}
	return idx;
}

Decl*
Deco::cxx_of_class(Decl *dco)
{
	if(dco->conv())
		return dco->conv();
	
	Id idc = dco->id();
	Scope psc = dco->scope();
	dco->unchain();
	dco->conv(dco);
	Decl *dc = forwref(Decl::dt_class, idc, psc);
	assert(dc);
	dco->conv(dc);
	dc->otype(dco->otype());
	if(!dco->defined())
		return dc;
	dcxx_init();
	
	Scope sc = newaggrscope(dc);

	// gen: private: static void *_itable[] = { ... };
	Decl *ditab = newdecl(_id__itable, sc, type_array(type_ptrto(&Type_void), NPOS));
	ditab->access(Decl::ac_private);
	ExprList eitab(ARRAY, LINE);
	ditab->vexpr(eitab);

	Itable itable;
	SymTab itsym(sc.strtab());
	itable.symtab = &itsym;
	itable.ditab = ditab;
	itable.eitab = &eitab;
	itable.sc = sc;
	itable.last = 0;

	for(Decl *ddi = dco->afirst(); ddi; ddi = ddi->next()) {
		if(ddi->dtype() == Decl::dt_impl) {
		    Decl *dci = impl_to_cxx(ddi, dc);
			Scope sci = getaggrscope(dci);
			for(Decl *dm = ddi->afirst(); dm; dm = dm->next()) {
				assert(dm->isfunc());
				char buf[MAXIDLEN];
				sprintf(buf, "%s__%s", ddi->cstr_id(), dm->cstr_id());
				Decl *d = dm->clone(sc, id(buf));
				FnInfo fi;
				assert(getfninfo(dm->rtype(), &fi));
				fi.sc.parent(sci);
				Decl *dn = dm->clone(sci);
				dm->conv(dn);
				Decl *diobj = sci.ufind(_id_iobj);
				assert(diobj);
				Expr ect = ExprCall(Expr(diobj), ExprList())
							.rtype(type_ptrto(dc->rtype()));
				Expr e = Expr(OP_PTRTO, ect, Expr(d));
				StmtList sl(fi.sc);
				sl += stmt_relaycall(e, &fi);
				dn->vexpr(sl);
			}
			addimpl(&itable, ddi, ddi->idecl());
			dci->defined(true);
		} else
			ddi->clone(sc); 
	}
	dc->defined(true);
	return dc;
}

Decl*
Deco::gen_conv_op(type_t ti, Scope sc, Decl *ddbind)
{
	// ( gen: operator ti() const return btr; { dbind(&btr); } )
	// gen: operator ti() const { ti btr; dbind(&btr); return btr; }
	sz buf[MAXIDLEN];
	buf[0] = '\0';
	Scope savescope = ccout.scope();
	ccout.scope(sc);
	type_string(ti, buf);
	ccout.scope(savescope);
	strpcat(buf, "operator ");
	Decl *dconv = newmethod(id(buf), sc, 0, true);
	StmtList sl2(dconv->fpscope());
	Decl *d = newdecl(_id_btr, dconv->fpscope(), ti);
	sl2 += Stmt(ExprDecl(d));
	sl2 += StmtCall(ddbind, elist(Expr(OP_REF, d)));
	sl2 += StmtReturn(d);
	dconv->vexpr(sl2).sinline(true);
	return dconv;
}

void
Deco::type_bases(Decl *dt, Scope sc, int *b)
{
	Scope sct = getaggrscope(dt);
	for(Decl *dbi = sct.get(ID_INHERIT); dbi; dbi = dbi->link()) {
		Decl *db = dbi->idecl();
		Decl *dcb = cxx_of_type(db);

		// gen: operator ti() { ti btr = *this; btr._itabp += #; return btr; }
		sz buf[MAXIDLEN];
		buf[0] = '\0';
		type_string(type_declref(dcb), buf);
		strpcat(buf, "operator ");
		Id bufid = id(buf);
		if(sc.uget(bufid))
			continue;
		Decl *dm = newmethod(bufid, sc, 0, true);
#if 0
		StmtList sl(dm->fpscope());
		Decl *d = newdecl(_id_btr, dm->fpscope(), type_declref(dcb));
		d->vexpr(Expr(OP_PTR, getthis(dm)));
		sl += Stmt(ExprDecl(d));
		sl += Stmt(Expr(OP_ASNADD, Expr(OP_DOT, d, sc.uget(_id__itabp)),
			Lit(&Type_int, *b)));
		sl += StmtReturn(d);
		dm->vexpr(sl).sinline(true);
#endif

		// type_t ti = type_declref(dcb);
		// gen_conv_op(ti, sc, ddbind);

		int n = 0;
		for(dm = db->afirst(); dm; dm = dm->next()) {
			if(dm->ismeth()) 
				type_mfn_to_cxx(dm, sc, *b, n++);
		}
		++*b;
		type_bases(db, sc, b);
	}
}

Decl*
Deco::cxx_of_type(Decl *dt)
{
	if(dt->conv()) 
		return dt->conv();

	Id idc = dt->id();
	Scope psc = dt->scope();
	dt->unchain();
	dt->conv(dt);
	Decl *dc = forwref(Decl::dt_class, idc, psc);
	assert(dc);
	dt->conv(dc);
	dc->otype(dt->otype());
	if(!dt->defined())
		return dc;
	dcxx_init();

	Scope sc = newaggrscope(dc);

	// gen: : public BTRef;
	inherit(dc, _d_BTRef, TDecl::ac_public);

	// member fns and implementation fns
	int n = 0;
	for(Decl *dm = dt->afirst(); dm; dm = dm->next()) {
		if(dm->dtype() == Decl::dt_impl) {
			// typeimpl_to_cxx(dm, dc);
		} else if(dm->ismeth()) {
			type_mfn_to_cxx(dm, sc, -1, n++);
			Id id_def = id_encode("default__", dm->cstr_id());
			Decl *dd = dm->clone(sc, id_def);
			dd->sinline(false);
		}
	
	}
	type_bases(dt, sc, &n);

	dc->defined(true);
	return dc;
}

Decl*
Deco::cxx_of_decl(Decl *d, bool fn)
{
	if(fn) {
		return d;
		if(d->isfunc() && d->initialized())  {
			if(idreal(d->id()))
		 		printf("rewriting %s\n", d->cstr_id());
			else
				printf("rewriting %p\n", d->id());
			Expr(d->vexpr()).dump();
			d->vexpr(expr_rewrite(d->vexpr(), rewrite_dcxx_to_cxx, null_expr));
			printf("after rewrite\n");
			Expr(d->vexpr()).dump();
		} else {
			d->ptype(convtype(d->ptype()));
			d->rtype(convtype(d->rtype()));
		}
		return d;
	} 
	switch(d->dtype()) {
	case Decl::dt_class:
	case Decl::dt_struct:
	case Decl::dt_union:
		if(!d->implements(0))
			return d;
		return cxx_of_class(d);
	case Decl::dt_type:
		return cxx_of_type(d);
	case Decl::dt_dclass:
		return d;
		// return cxx_of_dclass(d);
	/*
	case Decl::dt_default: {
		type_t pt = d->ptype();
		type_t npt = cxxtype(pt);
		if(pt != npt) {
			d->ptype(npt);
			d->rtype(realtype(npt));
		}
		} break; */
	default:
		return d;
	}
}

void
Deco::dcxx_to_cxx(Scope sc, bool fn)
{
	if(Decl *ds = sc.decl()) {
		for(Decl *dn, *d = ds->afirst(); d; d = dn) {
			dn = d->next();
			cxx_of_decl(d, fn);
		}
	} else 
		for(int n=0; Id id = sc.index(n); n++) 
			cxx_of_decl(sc.get(id), fn);
}


void
Deco::dcxx_to_cxx()
{
	for(ScopeList sl = _allscopes; Scope s = sl.head(); sl = sl.tail())
		dcxx_to_cxx(s, false);
	for(ScopeList sl = _allscopes; Scope s = sl.head(); sl = sl.tail())
		dcxx_to_cxx(s, true);
}


void
Deco::predefine()
{
	if(!dcxx())
		return;

	_id_instance = id("instance");
	_id_dtobj = id("dtobj");
	_id_dcobj = id("dcobj");
	_id_dataclass = id("dataclass");
	_id_dbind = id("dbind");
	_id_dcheck = id("dcheck");
	_id_dcbind = id("dcbind");
	_id_o = id("o");
	_id_btrp = id("btrp");
	_id__objp = id("_objp");
	_id__itabp = id("_itabp");
	_id__itable = id("_itable");
	_id_D_Type = id("D_Type");
	_id_D_Class = id("D_Class");
	_id_this_ref = id("this_ref");
	_id_this_r = id("this_r");
	_id_this_w = id("this_w");
	_id_this_rw = id("this_rw");
	_id_btr = id("btr");
	_id_copy = id("copy");
	_id_share = id("share");
	_id_iobj = id("iobj");
	_id__iobj = id("_iobj");
	_id_cobj = id("cobj");
	_id__cobj = id("_cobj");
	_id_any = id("any");
	_id__type = id("_type");
/*
	forwref(Decl::dt_class, (_id_BTRef = id("BTRef")));
	Decl *d = forwref(Decl::dt_class, (_id_Obj = id("Obj")));
	((Aggr*)d->rtype()->ptr())->scope(new SymTab(d->scope(), d));
	forwref(Decl::dt_class, (_id__DCBind = id("_DCBind")));
	forwref(Decl::dt_class, (_id_D_Type = id("D_Type")));
	forwref(Decl::dt_class, (_id_D_Class = id("D_Class")));
*/
}

Decl*
Deco::declfind(csz *name, Scope sc)
{
	if(!sc)
		sc = global();
	Decl *d = sc.ufind(id(name));
	if(!d) {
		error("`%s' not declared", name);
		_dcxxobj_initialized = false;
	}
	return d;
}

void
Deco::dcxx_init()
{
	if(_dcxx_initialized)
		return;
	if(declfind("__dchdr__")) {
		_d_BTRef = declfind("BTRef");
		Scope scbt = get_symtab(_d_BTRef);
		_d_bt__iobj = declfind("_iobj", scbt);
		_d_bt__cobj = declfind("_cobj", scbt);
		// _d_bt__DCBind = declfind("_DCBind", scbt);
		if(_d_BTRef) {
			_dcxx_initialized = true;
			return;
		} else
			abort("fatal error in include file <dchdr>");
	}
	abort("include file <dchdr> missing");
	return;
}

bool
Deco::dcxxobj_init()
{
	dcxx_init();
	if(_dcxxobj_initialized)
		return true;
	if(!declfind("__dcobj__"))
		return false;

	_dcxxobj_initialized = true;
	Scope scbt = get_symtab(_d_BTRef);
	_d_bt__DCBind = declfind("_DCBind", scbt);
	_d_bt__this_r = declfind("_this_r", scbt);
	_d_bt__this_w = declfind("_this_w", scbt);
	_d_bt__this_rw = declfind("_this_rw", scbt);
	_d_bt__this_ref = declfind("_this_ref", scbt);

	_d_Obj = declfind("Obj");
	Scope scobj = get_symtab(_d_Obj);
	_d_obj__init_alloc = declfind("_init_alloc", scobj);
	_d_obj__init_ref = declfind("_init_ref", scobj);
	_d_obj__this_r = declfind("_this_r", scobj);
	_d_obj__this_w = declfind("_this_w", scobj);
	_d_obj__this_rw = declfind("_this_rw", scobj);

	_d_D_Class = 0;
	_d_D_Class_dcobj = 0;
	_d_D_Type = 0;
	_d_D_Type_dcobj = 0;

	if(!_dcxxobj_initialized)
		abort("fatal error in include file <dcobj>");
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// I/O functionality

void
Deco::veprintf(csz *fmt, va_list args)
{
	vfprintf(stderr, fmt, args);
}



void
Deco::abort_exit()
{
	exit(1);
}


void
Deco::internal_error_exit()
{
	mprintf("", "Please send a full bug report to `shaggy@csh.rit.edu'");
	mprintf("", "(please read the BUGS file first for details)");
	exit(2);
}


bool
Deco::open_input(csz *fname, csz *cpp)
{
	_input_fd = fname ? open(fname, O_RDONLY) : 0;
	if(_input_fd >= 0 && cpp) {
		sz c[1024];
		sprintf(c, "%s %s", cpp, fname ? fname : "-");
		if(printexec)
			fprintf(stderr, "exec: %s\n", c);
		FILE *f = popen(c, "r");
		_input_fd = fileno(f);
	}
	_clex.source(fname ? fname : "");
	return _input_fd >= 0;
}


void
Deco::close_input()
{
	if(_input_fd)
		close(_input_fd);
}


void
Deco::get_input(bool split_token)
{
	if(isatty(_input_fd) && !split_token) 
		write(1, (complete() ? "> " : "? "), 2);
	int n = read(_input_fd, _buf, sizeof _buf);
	if(n <= 0)
		_clex.set_eof();
	else
		_clex.set_src(_buf, n);
}
 

int
main(int argc, sz **argv)
{
	char strbuf[4096];
	char bufbuf[4096];
	char namebuf[1024];
	sz cppopt[1024];
	bool stop_occ = false;
	bool stop_cpp = false;
	bool nolang = true;
	bool dumpscopes = false;
	bool dumpscopes0 = false;
	csz *cppcxx = DECOCPP " " DECODEFINES " " DECOINCLUDESCXX;
	csz *cppdcxx = DECOCPP " " DECODEFINES " " DECOINCLUDESDCXX;
//	csz *cxx = DECOCXX;
	int rv = 0;
	int nfiles = 0;

	cppopt[0] = '\0';
	
	CLex lex(strbuf, sizeof(strbuf));
	lex.set_buf(bufbuf, sizeof(bufbuf));
	lex.set_name(namebuf, sizeof(namebuf));
	lex.parse_keywords();
	lex.parse_space(false).parse_indent(false);
//	lex.parse_cpp();
//	lex.parse_eol();
//	lex.parse_indent();

	/* XXX - main loop needs restructuring.... */

	StrTab strtab;
	SymTab symtab(&strtab);
	Decl gdecl(0, 0);
	Scope(&symtab).decl(&gdecl);
	Deco deco(lex, strtab, symtab);
	CParse cparse(lex, deco);

	int argnum = 0;
	while(++argnum < argc) {
		csz *a = argv[argnum];

		if(a[0] == '-' && a[1]) {	// it's an option

			// Overall Options
			if(strcmp(a, "-x") == 0) {
				a = argv[++argnum];
				if(!a) {
					fprintf(stderr, "deco: missing arg\n");
					exit(1);
				}
				nolang = false;
				if(strcmp(a, "c") == 0)
					deco.parse_c();
				else if(strcmp(a, "c-hdr") == 0)
					deco.parse_h();
				else if(strcmp(a, "c-header") == 0)
					deco.parse_h();
				else if(strcmp(a, "c++") == 0)
					deco.parse_cxx();
				else if(strcmp(a, "c++hdr") == 0)
					deco.parse_hxx();
				else if(strcmp(a, "dc++") == 0)
					deco.parse_dcxx();
				else if(strcmp(a, "dc++hdr") == 0)
					deco.parse_dhxx();
				else if(strcmp(a, "none") == 0)
					nolang = true;
				else {
					fprintf(stderr,
					    "deco: unknown language `%s'\n",
						a);
					exit(1);
				}	
			}
			// -S
			else if(strcmp(a, "-E") == 0)
				stop_cpp = true;
			else if(strcmp(a, "-mcc") == 0)
				stop_occ = true;
			// -OC
			// -c
			else if(strcmp(a, "-np") == 0)
				cppcxx = cppdcxx = 0;
			// -o
			else if(strcmp(a, "-v") == 0) {
				printf("deco " deco_VER_STR " ("
					deco_VER_DATE ")\n");
				return 0;
			}

			// Preprocessor Options
			else if(strncmp(a, "-i", 2) == 0) {
				csz *b = argv[++argnum];
				if(!b) {
					fprintf(stderr, "deco: missing arg\n");
					exit(1);
				}
				sprintf(cppopt+strlen(cppopt), " %s %s", a, b);
			} else if(strncmp(a, "-I", 2) == 0
			     || strncmp(a, "-i", 2) == 0
			     || strncmp(a, "-D", 2) == 0
			     || strncmp(a, "-U", 2) == 0
			     || strncmp(a, "-A", 2) == 0
			     || strncmp(a, "-M", 2) == 0
			     || strncmp(a, "-nostdinc", 9) == 0
			     || strncmp(a, "-undef", 6) == 0
			     || strcmp(a, "-C") == 0
			     || strcmp(a, "-P") == 0
			     || strcmp(a, "-H") == 0
			     || strcmp(a, "-dM") == 0
			     || strcmp(a, "-dD") == 0
			     || strcmp(a, "-dN") == 0)
				sprintf(cppopt+strlen(cppopt), " %s", a);

			// Language and Parsing Options
			else if(strcmp(a, "-fsavecode") == 0)
				cparse.savecode(true);
			else if(strncmp(a, "-finclevel=", 11) == 0)
				deco.inclevel(atoi(a+11));

			// Formatting Options
			else if(strcmp(a, "-ffullscope") == 0)
				deco.ccout.fullscope();
			else if(strcmp(a, "-frealtypes") == 0)
				deco.realtypes();
			else if(strncmp(a, "-fwidth=", 8) == 0) 
				deco.ccout.width(atoi(a+8));
			else if(strcmp(a, "-finterface") == 0)
				deco.ccout.interface(true);
			else if(strcmp(a, "-fonlyinterface") == 0)
				deco.ccout.onlyinterface(true);

			// Warning Options
			else if(strcmp(a, "-w") == 0)
				deco.nowarnings();
			else if(strcmp(a, "-fsyntax-only") == 0)
				deco.nooutput(true);
			else if(strcmp(a, "-Wall") == 0) {
				// all of what warnings?
			}
			else if(strcmp(a, "-Werror") == 0)
				deco.warnerr();

			// Debugging Options
			// -g
			else if(strcmp(a, "-dy") == 0)
				cparse.yydebug(true);
			else if(strcmp(a, "-dA") == 0)
				debugaction = true;
			else if(strcmp(a, "-dT") == 0)
				deco.ccout.debugtypes();
			else if(strcmp(a, "-dC") == 0) 
				cparse.debug(true);
			else if(strcmp(a, "-dS") == 0)
				dumpscopes = true;
			else if(strcmp(a, "-dS0") == 0)
				dumpscopes0 = true;
			else if(strcmp(a, "-dX") == 0)
				printexec = true;
			else if(strcmp(a, "-dE") == 0)
				deco.ccout.debugexpr();
			else if(strcmp(a, "-da") == 0) {
				cparse.yydebug(true);
				debugaction = true;
				deco.ccout.debugtypes();
				deco.ccout.debugexpr();
				cparse.debug(true);
				dumpscopes = true;
			}
			else if(strcmp(a, "-ddef") == 0)
				deco._notedef = true;
			else if(strcmp(a, "-ddecl") == 0)
				deco._notedecl = true;
			// -save-temps

			// Code Generation Options

			// Optimization Options
		
			else
				fprintf(stderr,
					"deco: unknown option `%s'\n", a);
			continue;
		}

		++nfiles;
		csz *fname = strcmp(a, "-") == 0 ? 0 : a;
		if(nolang) {
		    deco.parse_dcxx();
		    if(fname) {
			ParsePath p(fname);
			csz *ext = p.get_ext(".");
			if(ext) {
				if(strcmp(ext, "c") == 0) 
					deco.parse_c();
				else if(strcmp(ext, "cc") == 0
						|| strcmp(ext, "C") == 0
						|| strcmp(ext, "c++") == 0
						|| strcmp(ext, "cxx") == 0)
					deco.parse_cxx();
				else if(strcmp(ext, "dh") == 0)
					deco.parse_dhxx();
				else if(strcmp(ext, "hh") == 0
						|| strcmp(ext, "H") == 0)
					deco.parse_hxx();
				else if(strcmp(ext, "h") == 0)
					deco.parse_h();
			}
		    }
		}
		sz cppbuf[4096];
		csz *cpp = deco.dcxx() ? cppdcxx : cppcxx;
		if(cpp) 
			sprintf(cppbuf, "%s%s%s%s", cpp, 
				" -D__dcplusplus ",
				deco.hdr() ? " -D__dcplusplus_hdr" : "",
				cppopt);
		if(!deco.open_input(fname, cpp ? cppbuf : 0)) {
			perror(fname ? fname : "stdin" );
			break;
		}
		if(stop_cpp) {
			while(!lex.eof()) {
				while(!lex.is_empty()) {
					printf("%c", lex.get_char());
					lex.next_char();
				} 
				deco.get_input(false);
			}
			continue;
		}
		deco.predefine();
		if(!cparse.parse()) 
			deco.mprintf("", "cannot recover from previous errors");
		if(dumpscopes0)
			deco.dump_scopes();
		if(deco.errors() == 0 && deco.dcxx())
			deco.dcxx_to_cxx();
		if(deco.warnings() > 0)
			deco.mprintf("", "%d warning%s",
				deco.warnings(), deco.warnings() == 1 ? "" : "s");
		if(dumpscopes)
			deco.dump_scopes();
		if(deco.errors() > 0) {
			deco.mprintf("", "%d error%s",
				deco.errors(), deco.errors() == 1 ? "" : "s");
			rv = 1;
		} else if (!deco.nooutput()) {
			if(deco.hdr() || deco.ccout.interface())
				deco.ccout.outmask(
					(1<<Decl::OT_HDR)|(1<<Decl::OT_CODE));
			else
				deco.ccout.outmask((1<<Decl::OT_FROMHDR)|
					(1<<Decl::OT_HDR)|(1<<Decl::OT_CODE)|
					(1<<Decl::OT_IMPLICIT)|(1<<Decl::OT_GEN));
			deco.ccout.oprintf("// this file was auto-generated"
				" by deco " deco_VER_STR " from the command:");
			deco.ccout.onl();
			deco.ccout.oprintf("//\t");
			for(int i=0; i<argc; i++)
				deco.ccout.oprintf("%s ", argv[i]);
			deco.ccout.onl(2); 
			deco.print_syms();
		}
		deco.close_input();
	}
	if(!nfiles) {
		fprintf(stderr, "deco: nothing to do\n");
		exit(1);
	} 
	return rv;
}

