/*  $Id: deco2.cc,v 1.2 1999/06/20 17:10:37 shaggy Exp $
 *      Dynamic Encapsulator of C++ Objects
 *      
 *      relative path: util/deco
 *  
 *  Copyright (c) 1996,1997 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define deco_VER_MAJOR	0
#define deco_VER_MINOR	12
#define deco_VER_PATCH	0
#define deco_VER_STR	"0.12"
#define deco_VER_DATE	"August 21, 1998"
#define deco_VER_NDATE	"08/21/98"

// #pragma Make Config_Needed += local
#pragma Make Includes_Banned += -nostdinc -nostdinc++
#pragma Make Flags_Banned += -static

#define DEBUG
#define CACTION_DEBUG

#define USE_STDLIB

#include <sys/types.h>
#include <defines>
#include <unistd.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cstdarg>
#include <fcntl.h>
#include <posix/assert.h>

#include <CObj>
#include <CLex>
#include <StrTab>
#include <SymTab>
#include <ParsePath>
#include <vCAction>
#include <CParse>
#include <CRep>
#include <COut>
#include <CAction>


bool debugaction = false;
bool printexec = false;

#define INTERNAL_ERROR  { internal_error(__FILE__, __LINE__, __FUNCTION__); \
			  for(;;); }

#include "decoconf.h"


// XXX - ideas:
//	- extend Item structure by making it a class that can
//	  be derived from and have methods added to
//	- extend Token by putting base definition in a file that can
//	  be included and then have more entries added to
//	- make sure specific/generic print routines are available for it 
//	  too

// XXX - at some point, any knowledge specific to certain data
// types should be removed from DECO, since it is supposed to be 
// a generic C++ pre-processor that generates code that can compile
// on any architecture. 


// Need to clean up use of Item by providing a local
// Item class that has a separate structure for each token ID,
// so that casting won't be necessary and the expected layout for
// each token is clear and can be checked for errors


// lexical shortcuts:
typedef vStrTab::Id Id;


Id _id_instance, _id_dtobj, _id_dcobj, _id_dataclass;
Id _id_dbind, _id_dcbind, _id_dcheck;
Id _id_o, _id_btrp;
Id _id_D_Type, _id_D_Class;
Id _id__objp;
Id _id__itabp;
Id _id__itable;
Id _id_this_ref;
Id _id_this_r;
Id _id_this_w;
Id _id_this_rw;
Id _id_btr;
Id _id_copy;
Id _id_share;

// declarations from dcobj.hh:
Decl *_d_BTRef;
Decl *_d_bt__DCBind;
Decl *_d_bt__this_r;
Decl *_d_bt__this_w;
Decl *_d_bt__this_rw;
Decl *_d_bt__this_ref;
Decl *_d_Obj;
Decl *_d_obj__init_alloc;
Decl *_d_obj__init_ref;
Decl *_d_obj__this_r;
Decl *_d_obj__this_w;
Decl *_d_obj__this_rw;
Decl *_d_obj__this_ref;

Decl *_d_D_Class;
Decl *_d_D_Class_dcobj;
Decl *_d_D_Type;
Decl *_d_D_Type_dcobj;


class Deco : public CAction {
private:
	int _input_fd;
	const int _bufsize = 4096;
	char _buf[_bufsize];


	bool _dcxx_initialized;

	Decl* declfind(csz *name, Scope sc = 0);
	bool dcxx_init();

public:
	Deco(CLex &clp, vStrTab &strtab, vSymTab &symtab);

	bool open_input(csz *fname = 0, csz *cpp = 0);
	void close_input();
	void get_input(bool split_token);
	void abort_exit();
	void internal_error_exit();
	void veprintf(csz *fmt, va_list args);

	void dcxx_to_cxx();
	void dcxx_to_cxx(Scope);
	Decl* cxx_of_dtype(Decl *);
	Decl* cxx_of_dclass(Decl *);
	Decl* cxx_of_decl(Decl *);
	Decl* dimpl_to_cxx(Decl *ddi, Decl *dc);
	void dtypedimpl_to_cxx(Decl *ddi, Decl *dc);
	void dclassdimpl_to_cxx(Decl *ddi, Decl *dc, Decl *ddat);
	void dtype_mfn_to_cxx(Decl *, Scope sdt, int n);
	Decl* gen_itable(Decl *ddi, Scope sci);
	void gen_dbind(Decl *dc, Decl *ditab, type_t ti);
	void gen_dcbind(Decl *dc, Decl *dci, Decl *ditab, Decl *ddcheck,
								type_t ti);
	void gen_binder(Decl *dc, Decl *dci, Decl *ditab, type_t ti);

	Decl* newmethod(Id id, Scope sc, type_t rt, bool is_const = false,
						int nargs = 0, Param * = 0);
	Expr expr_mfnadr(type_t newtype, Decl *d);
	ExprList copyargs(Func *f);

	void predefine();

	Decl* decltransplant(Decl *d);

	action EndAggr;
};


Deco::Deco(CLex &clp, vStrTab &strtab, vSymTab &symtab)
	 : CAction(clp, strtab, symtab)
{
	_dcxx_initialized = false;
}


void
Deco::EndAggr(Item *r, Item_ agsc, Item_ scope, Item_)
{
    PRINT_ACTION(r, agsc, scope, none);
	CAction::EndAggr(r, agsc, scope, none);
	Decl *d = uAggrId(r).decl;
	switch(d->dtype) {
//	case Decl::dt_dtype:
//		cxx_of_dtype(d);
//	case Decl::dt_dclass:
//		cxx_of_dclass(d);
	default: ;
	}	
}


/////////////////////////////////////////////////////////////////////////////
// code re-writing

class ExprRCAST : public Expr { public: ExprRCAST(type_t t, Expr e)
			: Expr(T(KW_RCAST), t, e) { set_rtype(realtype(t)); }};
class ExprSCAST : public Expr { public: ExprSCAST(type_t t, Expr e)
			: Expr(T(KW_SCAST), t, e) { set_rtype(realtype(t)); }};
class ExprCCAST : public Expr { public: ExprCCAST(type_t t, Expr e)
			: Expr(T(KW_CCAST), t, e) { set_rtype(realtype(t)); }};
class ExprCAST : public Expr { public: ExprCAST(type_t t, Expr e)
			: Expr(T(OP_CAST), t, e) { set_rtype(realtype(t)); }};


static Expr
expr_rewrite(Expr e, Expr (*rwf)(Expr, Expr), Expr p)
{
	Expr ne;
	if(!e)
		return e;
	// printf("eee=%p, %d, (%p, %p, %p) %p\n", (ExprNode*)e, e.op(), (ExprNode*)e.left(), (ExprNode*)e.mid(), (ExprNode*)e.right(), (ExprNode*)e.next());
	switch(e.op()) {
	case T(VAL_ID):
	case T(TOKENLIST):
	case T(LITERAL):
	case T(ID):
	case T(TYPE):
		ne = e;
		break;
	case T(EXPRLIST):
	case T(STMTLIST): {
		Expr eleft = expr_rewrite(e.left(), rwf, p);
		if(e.left().is(eleft))
			ne = e;
		else {
			Expr i = eleft;
			while(i.next())
				i = i.next();
			ne = Expr(e.op(), eleft, e.mid(), i);
		}
		} break;
	default: {
		Expr eleft = expr_rewrite(e.left(), rwf, p);
		Expr emid = expr_rewrite(e.mid(), rwf, p);
		Expr eright = expr_rewrite(e.right(), rwf, p);
		if(e.left().is(eleft) && e.mid().is(emid)
				&& e.right().is(eright))
			ne = e;
		else {
			ne = Expr(e.op(), eleft, emid, eright);
			ne.set_rtype(e.rtype());
		}}
	}
	Expr erw = rwf(ne, p);
	Expr enext = expr_rewrite(e.next(), rwf, p);
	if(!erw.next().is(enext)) {
		if(ne.is(e) || !ne.is(erw))
			ne = erw.dup();
		ne += enext;
	} else
		ne = erw;
	return ne;
}



/////////////////////////////////////////////////////////////////////////////
// DC++ to C++ conversions

// a lot of the following functions are rather grotesque
// hopefully, they will get trimmed down once better support functionality
// exists, since there's a lot of redundant code here
// The code here also employs "brute force" far too often.
// Something more elegant, but functionally equivalent, needs to replace it.
//
// Believe it or not, it's actually getting better already... 
// You should have seen it before some of this stuff got split
// into separate functions...

static ExprList
elist()
{
	return ExprList();
}

static ExprList
elist(Expr e1)
{
	return ExprList() += e1;
}

static ExprList
elist(Expr e1, Expr e2)
{
	return (ExprList() += e1) += e2;
}

/*
static ExprList
elist(Expr e1, Expr e2, Expr e3)
{
	return ((ExprList() += e1) += e2) += e3;
}
*/

ExprList
Deco::copyargs(Func *f)
{
	ExprList p;
	for(uint i=0; i<f->nparam(); i++) {
		Decl *pd = param_decl(f, i);
		if(!pd || !pd->id) {
			sz buf[10];
			sprintf(buf, "_a%d_", i);
			Id id = strid(buf);
			if(pd)
				declrename(pd, id);
			else {
				Param np;
				np.type = f->ptype(i);
				np.name = id;
				np.defval.token = NONE;
				pd = addparam(i, f->scope(), np);
			}
		}
		p += Expr(pd);
	}
	return p;
}

static Expr
rewrite_id_expr(Expr e, Expr p)
{
	if(e.op() == T(VAL_ID) && e.decl() == p.decl()) 
		return p.next();
	else
		return e;
}


static Expr
rewrite_dcxx_to_cxx(Expr e, Expr)
{	
	if(e.op() == T(OP_CONDASN)) {
		return Expr(T(OP_ASN), e.left(), e.right());
	} else if(e.op() == T(DECL)) {
#if 0	
		Decl *d = e.left().decl();
		if(d->condasn) {
			d->condasn = false;
			printf("d=%p\n", d);
			type_t t = Expr(d->v.expr).rtype();
			printf("t=%p\n", t);
			Aggr *aggr = dcast_Aggr(t);
			printf("Aggr=%p\n", aggr);
			Decl *dc = Scope(aggr->scope()).decl();
			printf("dc=%p\n", d);
			printf("conv=%d\n", dc->converted);
			aggr = dcast_Aggr(dc->next->rtype);
			Decl *ddcbind = Scope(aggr->scope()).ufind(_id_dcbind);
			printf("ddcbind=%p\n", ddcbind);
			Expr ef = Expr(T(OP_DOT), d->v.expr, ddcbind).set_rtype(ddcbind->rtype);
			printf("ok\n");
			d->v.expr = ExprCall(ef, elist(d)).set_rtype(d->rtype);
			printf("ok2\n");
		}
	} else if(Aggr *aggr = dcast_Aggr(e.rtype())) {
		Scope sc = Scope(aggr->scope());
		if(sc) {
			Decl *dc = sc.decl();
			if(dc->converted) 
				return Expr(e).set_rtype(dc->next->rtype);
		}
#endif
	}
	return e;	
}


// some little naming conventions, to try to keep things
// from getting too confusing:
// (yes, this borders on hungarian notation in some ways,
// which I personally can't stand, but given the size and
// complexity of some of these functions, I figured this
// would help avoid a bunch of errors-- of course, the better
// solution would be to break this up into enough smaller
// pieces so that one and two letter names would be sufficient)
//
//	d_XXX		Decl* for XXX
//	t_XXX		type_t for XXX
//	s_XXX		scope owned by XXX (*not* parent scope)
//	ps_XXX		parent scope for XXX
//	id_XXX		vStrTab::Id of XXX
//	f_XXX		Func* for XXX
//	m_XXX		Method* for XXX
//	e_XXX		Expr
//	el_XXX		Expression List
//	sl_XXX		Statement List
//	XXX_c		refers to class being constucted from dtype/dclass
//	XXX_dt		refers to dynamic type being converted
//	XXX_dc		refers to dynamic class being converted
//	XXX_dat		refers to dataclass (inside of dclass)
//	XXX_m		refers to current method
//	XXX_di		refers to dynamic implementation subclass
//	XXX_ci		refers to implementation subclass (converted)
//	XXX_dti		refers to dynamic type being implemented
//	XXX_cti		refers to class of dtype being implemented (converted)
//	XXX_itab	refers to itable[]
//	XXX_YYY		refers to YYY
//
// The usual convention applies about prefixing an identifier with _
// to show that it's a field.  Also, for local vars with limited
// scope, it's reasonable to drop the _ in the middle.

Decl*
Deco::newmethod(Id id, Scope s, type_t rt, bool is_const = false, int np, Param *param)
{
	type_t ptype[8];
	Scope pscope = newscope(s, 0);
	for(int i=0; i<np; i++) {
		ptype[i] = param[i].type;
		addparam(i, pscope, param[i]);
	}
	type_t psig = type_psig(np, ptype);
	type_t ot = s.decl()->rtype;
	if(is_const)
		ot = type_const(ot);
	Decl *d = newdecl(id, s, type_method(ot, type_func(rt, psig, pscope)));
	addthis(d);
	d->f.pscope = pscope;
	assert(d->is_func);
	assert(d->is_method);
	assert(!d->field);
	return d;
}


// return an expression that gets the address of a method function
// and converts it to the specified type
Expr
Deco::expr_mfnadr(type_t newtype, Decl *d)
{
	Expr emptr = Expr(OP_REF, Expr(d));

	Decl *dl = d->scope.uget(d->id)->link;
	if(dl && !dl->tagged())
		emptr = ExprSCAST(type_ptrto(d->rtype), emptr);
	return ExprCAST(newtype, emptr);
}


void
Deco::gen_dcbind(Decl *dc, Decl *dci, Decl *ditab, Decl *ddcheck, type_t ti)
{
	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;
	Scope sc = getaggrscope(dc);

	// gen: bool c::dcbind(::ti*) const
	// {
	//	btrp->_itabp = ti::_itable;
	//	btrp->_objp = const_cast<c*>(this);
	//	if(static_cast<ti*>(btrp)->dcheck())
	//		return true;
	//	else {
	//		btrp->_itabp = 0;
	//		btrp->_objp = 0;
	//		return false;
	//	};
	param[0].name = _id_btrp;
	param[0].type = type_ptrto(ti);
	Decl *ddcbind = newmethod(_id_dcbind, sc, &Type_bool, true, 1, param);
	StmtList sl(ddcbind->f.pscope);
	Decl *dbtrp = Scope(ddcbind->f.pscope).get(_id_btrp);
	Decl *dobjp = getaggrscope(_d_BTRef).ufind(_id__objp);
	Decl *ditabp = getaggrscope(_d_BTRef).ufind(_id__itabp);
	assert(dbtrp);
	assert(dobjp);
	assert(ditabp);
	sl += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(ditabp)),
		Expr(ditab)));
	sl += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(dobjp)),
		ExprRCAST(type_ptrto(type_declref(_d_Obj)),
			ExprCCAST(type_ptrto(type_declref(dc)),
				getthis(ddcbind)))));
	Expr ec = ExprCall(Expr(OP_PTRTO,
			ExprRCAST(type_ptrto(type_declref(dci)), dbtrp),
						ddcheck),
			ExprList());
	StmtList slelse(ddcbind->f.pscope);
	sl += StmtIfElse(ec, StmtReturn(Lit(&Type_bool, 1)), Stmt(slelse));
	slelse += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(ditabp)),
		Lit(&Type_int, 0)));
	slelse += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(dobjp)),
		Lit(&Type_int, 0)));
	slelse += StmtReturn(Lit(&Type_bool, 0));
	ddcbind->set_val(sl);

	// gen: ti c::dcbind(::ti&) const return btr; { dcbind(&btr); } 
	// gen: ti c::dcbind(::ti&) const {::ti btr; dcbind(&btr); return btr;}
	param[0].name = 0;
	param[0].type = type_refto(ti);
	Decl *ddcbind2 = newmethod(_id_dcbind, sc, ti, true, 1, param);
	StmtList sl2(ddcbind2->f.pscope);
	Decl *d = newdecl(_id_btr, ddcbind2->f.pscope, ti);
	sl2 += Stmt(ExprDecl(d));
	sl2 += StmtCall(ddcbind, elist(Expr(OP_REF, d)));
	sl2 += StmtReturn(d);
	ddcbind2->set_inline().set_val(sl2);
}

void
Deco::gen_dbind(Decl *dc, Decl *ditab, type_t ti)
{
	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;
	Scope sc = getaggrscope(dc);

	// gen: void dbind(ti *btrp) const 
	//	{ btrp->_objp = const_cast<c*>(this);
	// 	  btrp->_itabp = ti::_itable; }
	param[0].name = _id_btrp;
	param[0].type = type_ptrto(ti);
	Decl *ddbind = newmethod(_id_dbind, sc, &Type_void, true, 1, param);
	StmtList sl3(ddbind->f.pscope);
	Decl *dbtrp = Scope(ddbind->f.pscope).get(_id_btrp);
	Decl *dobjp = getaggrscope(_d_BTRef).ufind(_id__objp);
	Decl *ditabp = getaggrscope(_d_BTRef).ufind(_id__itabp);
	assert(dbtrp);
	assert(dobjp);
	assert(ditabp);
	sl3 += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(dobjp)),
		ExprRCAST(type_ptrto(type_declref(_d_Obj)),
			ExprCCAST(type_ptrto(type_declref(dc)),
				getthis(ddbind)))));
	sl3 += Stmt(Expr(OP_ASN, Expr(OP_PTRTO, Expr(dbtrp), Expr(ditabp)),
		Expr(ditab)));
	ddbind->set_val(sl3);

	// gen: operator ti() const return btr; { dbind(&btr); }
	// gen: operator ti() const { ti btr; dbind(&btr); return btr; }
	Decl *dconv;
	{
		sz buf[MAXIDLEN];
		buf[0] = '\0';
		Scope savescope = ccout.scope();
		ccout.scope(sc);
		type_string(ti, buf);
		ccout.scope(savescope);
		strpcat(buf, "operator ");
		dconv = newmethod(strid(buf), sc, 0, true);
	}
	StmtList sl2(dconv->f.pscope);
	Decl *d = newdecl(_id_btr, dconv->f.pscope, ti);
	sl2 += Stmt(ExprDecl(d));
	sl2 += StmtCall(ddbind, elist(Expr(OP_REF, d)));
	sl2 += StmtReturn(d);
	dconv->set_inline().set_val(sl2);
}


void
Deco::gen_binder(Decl *dc, Decl *dci, Decl *ditab, type_t ti)
{
	Scope sci = getaggrscope(dci);
	Decl *ddcheck = sci.uget(_id_dcheck);
	if(ddcheck)
		gen_dcbind(dc, dci, ditab, ddcheck, ti);
	else
		gen_dbind(dc, ditab, ti);
}

Decl*
Deco::dimpl_to_cxx(Decl *ddi, Decl *dc)
{
	if(!ddi->i.decl->defined)
		return 0;

	/*
	Decl *dci = forwref(Decl::dt_class, ddi->id, getaggrscope(dc));
	Scope sdi = getaggrscope(ddi);
	Scope sci = sdi;
	sdi = sdi.transplant();
	setaggrscope(dci, sci);
	*/
	Decl *dci = ddi;
	ddi = decltransplant(ddi);
	Scope sci = getaggrscope(dci);

	// gen: : public dti;
	assert(ddi->i.decl->from);
	Decl *dtic = ddi->i.decl;
	inherit(dci, dtic, TDecl::ac_public);

	// gen: private: inline ci() {}
	newmethod(ddi->id, sci, 0)->set_inline().set_val(StmtList())
				.set_access(Decl::ac_private);

	// gen: private: friend class c;
	newdecl(ID_FRIEND, sci, type_declref(dc))->storage = Decl::sc_friend;
	
	return ddi;
}

Decl*
Deco::gen_itable(Decl *ddi, Scope sci)
{
	// gen: private: static void *_itable[];
	Decl *ditab = newdecl(_id__itable, sci, type_array(type_ptrto(&Type_void), NPOS));
	ditab->set_access(Decl::ac_private);

	// gen: static void * c::ci::_itable[] = 
	//		{ 0, static_cast<void*>(fn), ... }
	ExprList eitab(ARRAY, LINE);
	eitab += Expr(Lit(&Type_int, 0));
	Decl *dtic = ddi->i.decl;
	Decl *dti = dtic->from;
	assert(dti);
	for(Decl *dtm = dti->a.first; dtm; dtm = dtm->next) {
		if(!dtm->is_method)
			continue;
		Decl *d = sci.uget(dtm->id);
		printf("d=%p d->scope=%p sci=%p\n",
			d, (vSymTab*)d->scope, (vSymTab*)sci);
		if(!d || d->scope != sci) {
			sz buf[MAXIDLEN];
			sprintf(buf, "default__%s", cstr(dtm->id));
			d = getaggrscope(dtic).uget(strid(buf));
		}
		if(!d)
			error("can't find definition for %s", cstr(dtm->id));
		eitab += expr_mfnadr(type_ptrto(&Type_void), d);
	}
	ditab->set_val(eitab);

	return ditab;
}


void
Deco::dtypedimpl_to_cxx(Decl *ddi, Decl *dc)
{
	Decl *dci = ddi;
	if(!(ddi = dimpl_to_cxx(ddi, dc)))
		return;
	Scope sci = getaggrscope(dci);

	// gen: c * this_ref() const { return static_cast<c*>(_this_ref()); }
	type_t rt = type_ptrto(type_declref(dc));
	Decl *dthisr = newmethod(_id_this_ref, sci, rt, true);
	StmtList sl;
	sl += StmtReturn(ExprRCAST(rt, ExprCall(Expr(_d_bt__this_ref),
							ExprList())));
	Expr ethiscall = ExprCall(Expr(dthisr), ExprList())
			.set_rtype(realtype(rt));
	dthisr->set_inline().set_val(sl);

	// gen: ti c::dcbind(::ti&) const return btr; { dcbind(&btr); } 
	// gen: ti c::dcbind(::ti&) const {::ti btr; dcbind(&btr); return btr;}
	
	for(Decl *dm = ddi->a.first; dm; dm = dm->next) {
		Decl *dp = decltransplant(dm);
		if(dp->is_method) {
			Method *meth = dcast_Method(dp->rtype);
			Func *func = dcast_Func(meth->ftype());
			Scope scf = func->scope();
			scf.parent(sci);
			dm->v.expr = expr_rewrite(dp->v.expr, rewrite_id_expr,
					Expr(getthis(dm)) += ethiscall);
		}
	}

	Decl *ditab = gen_itable(ddi, sci);

	// XXX - loop through type and add fns
	// (should perhaps use "struct itable" to provide ptr types?)

	gen_binder(dc, dci, ditab, type_declref(ddi->i.decl));

	dci->defined = true;
}

void
Deco::dclassdimpl_to_cxx(Decl *ddi, Decl *dc, Decl *ddat)
{
	Decl *dci = ddi;
	if(!(ddi = dimpl_to_cxx(ddi, dc)))
		return;
	Scope sci = getaggrscope(dci);

	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;

	// gen: dat * this_rw() const { return static_cast<dat*>(_this_rw()); }
	type_t rt = type_ptrto(type_declref(ddat));
	Decl *dthisrw = newmethod(_id_this_rw, sci, rt, true);
	StmtList sl;
	sl += StmtReturn(ExprRCAST(rt, ExprCall(Expr(_d_bt__this_rw),
							ExprList())));
	Expr ethiscall = ExprCall(Expr(dthisrw), ExprList())
			.set_rtype(realtype(rt));
	dthisrw->set_inline().set_val(sl);

	for(Decl *dm = ddi->a.first; dm; dm = dm->next) {
		Decl *dp = decltransplant(dm);
		if(dp->is_method) {
			if(!dp->initialized)
				continue;
			Method *meth = dcast_Method(dp->rtype);
			Func *func = dcast_Func(meth->ftype());
			Scope scf = func->scope();
			scf.parent(sci);
			if(Expr(dp->v.expr).op() == VAL_ID) {
				Decl *df = Expr(dp->v.expr).decl();
				assert(df);
				ExprList elp = copyargs(func);
				Expr ect = ExprCall(Expr(dthisrw), ExprList())
							.set_rtype(type_ptrto(ddat->rtype));
				Expr e = ExprCall(Expr(OP_PTRTO, ect, Expr(df)), elp);
				StmtList sl(scf);
				if(func->rtype() && func->rtype() != &Type_void)
					sl += StmtReturn(e);
				else
					sl += Stmt(e);
				dm->v.expr = sl;
			} else 
				dm->v.expr = expr_rewrite(dp->v.expr, rewrite_id_expr,
					Expr(getthis(dm)) += ethiscall);
		}
		dm->scope = sci;
	}

	Decl *ditab = gen_itable(ddi, sci);

	// XXX - loop through type and add fns
	// (should perhaps use "struct itable" to provide ptr types?)

	gen_binder(dc, dci, ditab, type_declref(ddi->i.decl));

	dci->defined = true;
}

void
Deco::dtype_mfn_to_cxx(Decl *dm, Scope sdt, int n)
{
	printf("dm=%p\n", dm);
	dm->next = 0;
	dm->link = 0;
	Decl *dp = new Decl;
	*dp = *dm;

	dm->from = dp;
	decladd(dm);
	chaindecl(dm);

	dp->scope = sdt;
	decladd(dp);
	chaindecl(dp);

	if(dp->initialized) {
		sz buf[MAXIDLEN];
		sprintf(buf, "default__%s", cstr(dp->id));
		Decl *dn = new Decl;
		*dn = *dp;
		dn->scope = dm->scope;
		dn->id = strid(buf);
		dn->where = -1;
		dn->link = 0;
		dn->next = 0;
		decladd(dn);
		chaindecl(dn);
	}
	Method *m = dcast_Method(dp->rtype);
	assert(m);
	Func *f = dcast_Func(m->ftype());
	assert(f);
	dm->defined = true;
	dm->initialized = true;
	StmtList sl(f->scope());
	ExprList elp = copyargs(f);
	Expr e = Expr(OP_INDEX, dm->scope.uget(_id__itabp),
				Lit(&Type_int, n));
	e = ExprRCAST(type_ptrto(dm->ptype), e);
	e = Expr(OP_MPTRTO, Expr(getthis(dm)), e);
	e = ExprCall(e, elp);
	if(f->rtype() == &Type_void)
		sl += Stmt(e);
	else
		sl += StmtReturn(e);
	dm->set_inline().set_val(sl);
}

Decl*
Deco::cxx_of_dtype(Decl *ddt)
{
	printf("converting %s (%p)\n", cstr(ddt->id), ddt->from);
	assert(ddt->dtype == Decl::dt_dtype);

	Decl *d;

	Id idc = ddt->id;
	Decl *dc = ddt;
	ddt = decltransplant(ddt);
	assert(ddt);
	dc->dtype = Decl::dt_class;
	if(!ddt->defined)
		return dc;
	if(!dcxx_init()) {
		error("must include <dcobj> to define dynamic type '%s'", cstr(idc));
		return 0;
	}
	
	Scope sc = getaggrscope(dc);
	assert(sc);
	Scope sdt = getaggrscope(ddt);
	assert(sdt);

	// type_t tdc = dc->rtype;
	// type_t tcdc = type_const(tdc);

	// gen: : public BTRef;
	inherit(dc, _d_BTRef, TDecl::ac_public);

	// gen:  static const D_Type dtobj;
	if(!_d_D_Type) {
		Decl *d = global().tget(_id_D_Type);
		if(d && d->dtype == Decl::dt_dclass)
			d = cxx_of_dclass(d);
		if(!d || d->dtype != Decl::dt_class) {
			error("`dclass D_Type' not declared");
			return 0;
		}
		_d_D_Type = d;
	}
	Decl *ddtobj = newdecl(_id_dtobj, sc,
				type_const(type_declref(_d_D_Type)));
	if(!_d_D_Type_dcobj) {
		Scope s = getaggrscope(_d_D_Type);
		if(!s) {
			error("`dclass D_Type' not defined");
			return 0;
		}
		if(!(_d_D_Type_dcobj = s.uget(_id_dcobj))) {
			error("`D_Type::dcobj' not declared");
			return 0;
		}
	}
	ddtobj->set_val(elist(_d_D_Type_dcobj, Lit(&Type_int, 0)));

	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;

	// member fns and implementation fns
	int n = 0;
	Decl *next = ddt->a.first;
	ddt->a.first = 0;
	ddt->a.tsize = 0;
	for(Decl *dm = next; dm; dm = next) {
		next = dm->next;
		printf("dm=%p next=%p\n", dm, next);
		if(dm->dtype == Decl::dt_dimpl) {
			dtypedimpl_to_cxx(dm, dc);
		} else if(dm->is_method)
			dtype_mfn_to_cxx(dm, sdt, ++n);
	}

	// gen: inline c(Obj &o, _DCBind) : BTRef(dtobj, &o) {}
	param[0].type = type_refto(type_declref(_d_Obj));
	param[0].name = _id_o;
	param[1].type = type_declref(_d_bt__DCBind);
	param[1].name = 0;
	d = newmethod(idc, sc, 0, false, 2, param);
	StmtList slctor(d->f.pscope);
	slctor += StmtCall(Expr(_d_BTRef), elist(Expr(ddtobj),
			Expr(OP_REF, Expr(Scope(d->f.pscope).uget(_id_o)))));
	d->set_inline().set_val(ExprCtor(slctor, StmtList()));

	// gen: inline c() : BTRef(dtobj) { _objp = 0; }
	d = newmethod(idc, sc, 0);
	slctor = StmtList(d->f.pscope);
	StmtList sl(d->f.pscope);
	slctor += StmtCall(Expr(_d_BTRef), elist(Expr(ddtobj)));
	Decl *dobjp = sc.ufind(_id__objp);
	sl += Stmt(Expr(OP_ASN, Expr(dobjp), Lit(&Type_int, 0)));
	d->set_inline().set_val(ExprCtor(slctor, sl));
	
	// gen: class instance : Obj {};
	d = forwref(Decl::dt_class, _id_instance, sc);
	newaggrscope(d);
	inherit(d, _d_Obj, TDecl::ac_public);
	d->defined = true;

	dc->defined = true;
	return dc;
}


// move a Decl to a new place (along with any associated scopes)
// and then clean out the old place, leaving info such
// as dtype, storage, etc, but clearing out the scope
// (the scope gets transplanted as well)

Decl*
Deco::decltransplant(Decl *d)
{
	Decl *nd = new Decl;
	*nd = *d;
	type_t t = d->rtype;
	if(Aggr *a = dcast_Aggr(t)) {
		if(vSymTab *st = a->scope()) {
			vSymTab *nst = st->transplant();
			Scope(nst).decl(nd);
			nd->ptype = type_aggr(d->id, nst);
		}
		d->a.first = 0;
		d->a.tsize = 0;
	} else
		d->defined = false;
	d->initialized = false;
	d->from = nd;
	nd->next = 0;
	nd->link = 0;
	// XXX - perhaps we should clear/change more fields in nd?
	return nd;
}

Decl*
Deco::cxx_of_dclass(Decl *ddc)
{
	printf("converting dclass %s (%p)\n", cstr(ddc->id), ddc->from);
	assert(ddc->dtype == Decl::dt_dclass);

	Decl *d;
	Param param[2];
	param[0].clear();
	param[0].defval = NONE;
	param[1].clear();
	param[1].defval = NONE;

	Id idc = ddc->id;
	Decl *dc = ddc;
	ddc = decltransplant(ddc);
	assert(ddc);
	dc->dtype = Decl::dt_class;
	if(!ddc->defined)
		return dc;
	if(!dcxx_init()) {
		error("must include <dcobj> to define dynamic class '%s'", cstr(idc));
		return 0;
	}
	
	Scope sc = getaggrscope(dc);
	type_t tdc = dc->rtype;
	type_t tconstdc = type_const(tdc);

	// gen: : public Obj;
	inherit(dc, _d_Obj, TDecl::ac_public);

	// gen:  static const D_Class dcobj;
	if(!_d_D_Class) {
		Decl *d = global().tget(_id_D_Class);
		if(d && d->dtype == Decl::dt_dclass)
			d = cxx_of_dclass(d);
		if(!d || d->dtype != Decl::dt_class) {
			error("`dclass D_Class' not declared");
			return 0;
		}
		_d_D_Class = d;
	}
	Decl *ddcobj = newdecl(_id_dcobj, sc,
			type_const(type_declref(_d_D_Class)));
	ddcobj->otype = ddc->otype;
	if(!_d_D_Class_dcobj) {
		Scope s = getaggrscope(_d_D_Class);
		if(!s) {
			error("`dclass D_Class' not defined");
			return 0;
		}
		if(!(_d_D_Class_dcobj = s.uget(_id_dcobj))) {
			error("`D_Class::dcobj' not declared");
			return 0;
		}
	}
	ddcobj->set_val(elist(_d_D_Class_dcobj, Lit(&Type_int, 0)));

	// gen: class dataclass : bases { newmembers; }
	Decl *ddat = forwref(Decl::dt_class, _id_dataclass, sc);
	Scope sdat = newaggrscope(ddat);
	Scope sdc = getaggrscope(ddc);
	int nbase = 0;
	Decl *dbase = 0;
	bool simple = true;
	for(Decl *dbi = sdc.get(ID_INHERIT); dbi; dbi = dbi->link) {
		Decl *db = dbi->i.decl;
		// XXX - check for dclass inheritance
		// error for dtype inheritance
		inherit(ddat, db, TDecl::ac_public);
		++nbase;
		dbase = db;
	}

	// transfer constructors from dc to c
	// transfer fields from dc to dat
	for(Decl *dm = ddc->a.first; dm; dm = dm->next) {
		if(dm->dtype == Decl::dt_dimpl) {
			dclassdimpl_to_cxx(dm, dc, ddat);
		} else if(dm->is_method) {
			d = new Decl;
			*d = *dm;
			d->where = -1;
			d->link = 0;
			d->next = 0;
			d->scope = sc;
			decladd(d);
			chaindecl(d);
		} else {
			d = new Decl;
			*d = *dm;
			d->where = -1;
			d->link = 0;
			d->next = 0;
			d->scope = sdat;
			decladd(d);
			chaindecl(d);
			simple = false;
		}
	}
	if(simple && nbase == 0)
		warning("implementation has no local data");
	if(nbase != 1)
		simple = false;

	if(simple) {
		// gen: dataclass(const base &copy) : base(copy) {}
		param[0].name = _id_copy;
		param[0].type = type_refto(type_const(dbase->rtype));
		d = newmethod(_id_dataclass, sdat, 0, false, 1, param);
		StmtList slctor(d->f.pscope);
		slctor += StmtCall(Expr(dbase),
			elist(Scope(d->f.pscope).uget(_id_copy)));
		d->set_inline().set_val(ExprCtor(slctor, StmtList()));

		// copy other constructors 
		for(Decl *dm = dbase->a.first; dm; dm = dm->next) {
			if(dm->is_method) {
				Method *meth = dcast_Method(dm->rtype);
				Func *f = dcast_Func(meth->ftype());
				if(f->rtype())
					continue;
				d = new Decl;
				*d = *dm;
				d->id = _id_dataclass;
				d->scope = sdat;
				d->where = -1;
				d->link = 0;
				d->next = 0;
				ExprList p = copyargs(f);
				StmtList slctor(f->scope());
				slctor += StmtCall(Expr(dbase), p);
				d->set_inline().set_val(ExprCtor(slctor,
							StmtList()));
				decladd(d);
				chaindecl(d);
			} 
		}

		// gen: dc(base *share) { _init_ref(dcobj, share); }
		param[0].name = _id_share;
		param[0].type = type_ptrto(dbase->rtype);
		d = newmethod(idc, sc, 0, false, 1, param);
		StmtList sl(d->f.pscope);
		sl += StmtCall(_d_obj__init_ref,
			elist(ddcobj, Scope(d->f.pscope).get(_id_share)));
		d->set_inline().set_val(sl);
	}
	ddat->defined = true;

	// copy dataclass constructors to dclass
	for(Decl *dm = ddat->a.first; dm; dm = dm->next) {
		if(dm->is_method) {
			Method *meth = dcast_Method(dm->rtype);
			Func *f = dcast_Func(meth->ftype());
			if(f->rtype())
				continue;
			d = new Decl;
			*d = *dm;
			d->id = idc;
			d->scope = sc;
			d->where = -1;
			d->link = 0;
			d->next = 0;
			ExprList elp = copyargs(f);
			StmtList sl(f->scope());
			Expr ea = ExprCall(_d_obj__init_alloc, 
				elist(ddcobj, Expr(OP_TSIZEOF, ddat)));
			sl += Stmt(Expr(OP_NEW, Expr(ddat->rtype), Expr(elp), ea));
			d->set_inline().set_val(sl);
			decladd(d);
			chaindecl(d);
		}
	}	

	// gen: const dataclass* this_r() const
	//	{ return static_cast<const dataclass*>(_this_r()); }
	Scope pscope = newscope(sc, 0);
	type_t psig = type_psig(0, 0);
	type_t rt = type_ptrto(type_const(type_declref(ddat)));
	StmtList s;
	newdecl(_id_this_r, sc, type_method(tconstdc,
					type_func(rt, psig, pscope)))
		->set_inline().set_val(s);
	s += StmtReturn(ExprRCAST(rt, ExprCall(Expr(_d_obj__this_r),
							ExprList())));

	// gen: dataclass* this_rw() { return static_cast<dataclass*>(_this_rw()); }
	pscope = newscope(sc, 0);
	psig = type_psig(0, 0);
	rt = type_ptrto(type_declref(ddat));
	StmtList s2;
	newdecl(_id_this_rw, sc, type_method(tdc, type_func(rt, psig, pscope)))
		->set_inline().set_val(s2);
	s2 += StmtReturn(ExprRCAST(rt, ExprCall(Expr(_d_obj__this_rw),
							ExprList())));

	dc->defined = true;
	return dc;
}

Decl*
Deco::cxx_of_decl(Decl *d)
{
	switch(d->dtype) {
	case Decl::dt_dtype:
		return cxx_of_dtype(d);
	case Decl::dt_dclass:
		return cxx_of_dclass(d);
	/*
	case Decl::dt_default: {
		type_t pt = d->ptype;
		type_t npt = cxxtype(pt);
		if(pt != npt) {
			d->ptype = npt;
			d->rtype = realtype(npt);
		}
		} break; */
	default:
		if(d->is_func && d->initialized)  {
			/*
			if(idreal(d->id))
				printf("rewriting %s\n", cstr(d->id));
			else
				printf("rewriting %p\n", d->id);
			*/
			d->v.expr = expr_rewrite(d->v.expr, rewrite_dcxx_to_cxx, null_expr);
		}
		return 0;
	}
}

void
Deco::dcxx_to_cxx(Scope sc)
{
	if(Decl *ds = sc.decl()) {
		for(Decl *dn, *d = ds->a.first; d; d = dn) {
			dn = d->next;
			cxx_of_decl(d);
		}
	} else 
		for(int n=0; Id id = sc.index(n); n++) 
			cxx_of_decl(sc.get(id));
}


void
Deco::dcxx_to_cxx()
{
	for(ScopeList sl = _allscopes; Scope s = sl.head(); sl = sl.tail())
		dcxx_to_cxx(s);
}


void
Deco::predefine()
{
	if(!dcxx())
		return;

	_id_instance = strid("instance");
	_id_dtobj = strid("dtobj");
	_id_dcobj = strid("dcobj");
	_id_dataclass = strid("dataclass");
	_id_dbind = strid("dbind");
	_id_dcheck = strid("dcheck");
	_id_dcbind = strid("dcbind");
	_id_o = strid("o");
	_id_btrp = strid("btrp");
	_id__objp = strid("_objp");
	_id__itabp = strid("_itabp");
	_id__itable = strid("_itable");
	_id_D_Type = strid("D_Type");
	_id_D_Class = strid("D_Class");
	_id_this_ref = strid("this_ref");
	_id_this_r = strid("this_r");
	_id_this_w = strid("this_w");
	_id_this_rw = strid("this_rw");
	_id_btr = strid("btr");
	_id_copy = strid("copy");
	_id_share = strid("share");
/*
	forwref(Decl::dt_class, (_id_BTRef = strid("BTRef")));
	Decl *d = forwref(Decl::dt_class, (_id_Obj = strid("Obj")));
	((Aggr*)d->rtype->ptr())->set_scope(new SymTab(d->scope, d));
	forwref(Decl::dt_class, (_id__DCBind = strid("_DCBind")));
	forwref(Decl::dt_class, (_id_D_Type = strid("D_Type")));
	forwref(Decl::dt_class, (_id_D_Class = strid("D_Class")));
*/
}

Decl*
Deco::declfind(csz *name, Scope sc)
{
	if(!sc)
		sc = global();
	Decl *d = sc.ufind(strid(name));
	if(!d) {
		error("`%s' not declared", name);
		_dcxx_initialized = false;
	}
	return d;
}

bool
Deco::dcxx_init()
{
	if(_dcxx_initialized)
		return true;
	if(!declfind("__dcobj__"))
		return false;

	_dcxx_initialized = true;
	_d_BTRef = declfind("BTRef");
	Scope scbt = get_symtab(_d_BTRef);
	_d_bt__DCBind = declfind("_DCBind", scbt);
	_d_bt__this_r = declfind("_this_r", scbt);
	_d_bt__this_w = declfind("_this_w", scbt);
	_d_bt__this_rw = declfind("_this_rw", scbt);
	_d_bt__this_ref = declfind("_this_ref", scbt);

	_d_Obj = declfind("Obj");
	Scope scobj = get_symtab(_d_Obj);
	_d_obj__init_alloc = declfind("_init_alloc", scobj);
	_d_obj__init_ref = declfind("_init_ref", scobj);
	_d_obj__this_r = declfind("_this_r", scobj);
	_d_obj__this_w = declfind("_this_w", scobj);
	_d_obj__this_rw = declfind("_this_rw", scobj);

	_d_D_Class = 0;
	_d_D_Class_dcobj = 0;
	_d_D_Type = 0;
	_d_D_Type_dcobj = 0;

	if(!_dcxx_initialized)
		abort("fatal error in include file <dcobj>");
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// I/O functionality

void
Deco::veprintf(csz *fmt, va_list args)
{
	vfprintf(stderr, fmt, args);
}



void
Deco::abort_exit()
{
	exit(1);
}


void
Deco::internal_error_exit()
{
	mprintf("", "Please send a full bug report to `shaggy@csh.rit.edu'");
	mprintf("", "(please read the BUGS file first for details)");
	exit(2);
}


bool
Deco::open_input(csz *fname, csz *cpp)
{
	_input_fd = fname ? open(fname, O_RDONLY) : 0;
	if(_input_fd >= 0 && cpp) {
		sz c[1024];
		sprintf(c, "%s %s", cpp, fname ? fname : "-");
		if(printexec)
			fprintf(stderr, "exec: %s\n", c);
		FILE *f = popen(c, "r");
		_input_fd = fileno(f);
	}
	_clex.source(fname ? fname : "");
	return _input_fd >= 0;
}


void
Deco::close_input()
{
	if(_input_fd)
		close(_input_fd);
}


void
Deco::get_input(bool split_token)
{
	if(isatty(_input_fd) && !split_token) 
		write(1, (complete() ? "> " : "? "), 2);
	int n = read(_input_fd, _buf, _bufsize);
	if(n <= 0)
		_clex.set_eof();
	else
		_clex.set_src(_buf, n);
}
 

int
main(int argc, sz **argv)
{
	char strbuf[4096];
	char bufbuf[4096];
	char namebuf[1024];
	sz cppopt[1024];
	bool stop_occ = false;
	bool stop_cpp = false;
	bool nolang = true;
	bool dumpscopes = false;
	bool dumpscopes0 = false;
	csz *cpp = DECOCPP " " DECODEFINES " " DECOINCLUDES;
//	csz *cxx = DECOCXX;
	int rv = 0;
	int nfiles = 0;

	cppopt[0] = '\0';
	
	CLex lex(strbuf, sizeof(strbuf));
	lex.set_buf(bufbuf, sizeof(bufbuf));
	lex.set_name(namebuf, sizeof(namebuf));
	lex.parse_keywords();
	lex.parse_space(false).parse_indent(false);
//	lex.parse_cpp();
//	lex.parse_eol();
//	lex.parse_indent();

	/* XXX - main loop needs restructuring.... */

	StrTab strtab;
	SymTab symtab(&strtab);
	Decl gdecl;
	gdecl.clear();
	Scope(&symtab).decl(&gdecl);
	Deco deco(lex, strtab, symtab);
	CParse cparse(lex, deco);

	int argnum = 0;
	while(++argnum < argc) {
		csz *a = argv[argnum];

		if(a[0] == '-' && a[1]) {	// it's an option

			// Overall Options
			if(strcmp(a, "-x") == 0) {
				a = argv[++argnum];
				if(!a) {
					fprintf(stderr, "deco: missing arg\n");
					exit(1);
				}
				nolang = false;
				if(strcmp(a, "c") == 0)
					deco.parse_c();
				else if(strcmp(a, "c-hdr") == 0)
					deco.parse_h();
				else if(strcmp(a, "c-header") == 0)
					deco.parse_h();
				else if(strcmp(a, "c++") == 0)
					deco.parse_cxx();
				else if(strcmp(a, "c++hdr") == 0)
					deco.parse_hxx();
				else if(strcmp(a, "dc++") == 0)
					deco.parse_dcxx();
				else if(strcmp(a, "dc++hdr") == 0)
					deco.parse_dhxx();
				else if(strcmp(a, "none") == 0)
					nolang = true;
				else {
					fprintf(stderr,
					    "deco: unknown language `%s'\n",
						a);
					exit(1);
				}	
			}
			// -S
			else if(strcmp(a, "-E") == 0)
				stop_cpp = true;
			else if(strcmp(a, "-mcc") == 0)
				stop_occ = true;
			// -OC
			// -c
			else if(strcmp(a, "-np") == 0)
				cpp = 0;
			// -o
			else if(strcmp(a, "-v") == 0) {
				printf("deco " deco_VER_STR " ("
					deco_VER_DATE ")\n");
				return 0;
			}

			// Preprocessor Options
			else if(strncmp(a, "-i", 2) == 0) {
				csz *b = argv[++argnum];
				if(!b) {
					fprintf(stderr, "deco: missing arg\n");
					exit(1);
				}
				sprintf(cppopt+strlen(cppopt), " %s %s", a, b);
			} else if(strncmp(a, "-I", 2) == 0
			     || strncmp(a, "-i", 2) == 0
			     || strncmp(a, "-D", 2) == 0
			     || strncmp(a, "-U", 2) == 0
			     || strncmp(a, "-A", 2) == 0
			     || strncmp(a, "-M", 2) == 0
			     || strncmp(a, "-nostdinc", 9) == 0
			     || strncmp(a, "-undef", 6) == 0
			     || strcmp(a, "-C") == 0
			     || strcmp(a, "-P") == 0
			     || strcmp(a, "-H") == 0
			     || strcmp(a, "-dM") == 0
			     || strcmp(a, "-dD") == 0
			     || strcmp(a, "-dN") == 0)
				sprintf(cppopt+strlen(cppopt), " %s", a);

			// Language and Parsing Options
			else if(strcmp(a, "-fsavecode") == 0)
				cparse.savecode(true);
			else if(strncmp(a, "-finclevel=", 11) == 0)
				deco.inclevel(atoi(a+11));

			// Formatting Options
			else if(strcmp(a, "-ffullscope") == 0)
				deco.ccout.fullscope();
			else if(strcmp(a, "-frealtypes") == 0)
				deco.realtypes();
			else if(strncmp(a, "-fwidth=", 8) == 0) 
				deco.ccout.width(atoi(a+8));
			else if(strcmp(a, "-finterface") == 0)
				deco.ccout.interface(true);
			else if(strcmp(a, "-fonlyinterface") == 0)
				deco.ccout.onlyinterface(true);

			// Warning Options
			else if(strcmp(a, "-w") == 0)
				deco.nowarnings();
			else if(strcmp(a, "-fsyntax-only") == 0)
				deco.nooutput(true);
			else if(strcmp(a, "-Wall") == 0) {
				// all of what warnings?
			}
			else if(strcmp(a, "-Werror") == 0)
				deco.warnerr();

			// Debugging Options
			// -g
			else if(strcmp(a, "-dy") == 0)
				cparse.yydebug(true);
			else if(strcmp(a, "-dA") == 0)
				debugaction = true;
			else if(strcmp(a, "-dT") == 0)
				deco.ccout.debugtypes();
			else if(strcmp(a, "-dC") == 0) 
				cparse.debug(true);
			else if(strcmp(a, "-dS") == 0)
				dumpscopes = true;
			else if(strcmp(a, "-dS0") == 0)
				dumpscopes0 = true;
			else if(strcmp(a, "-dX") == 0)
				printexec = true;
			else if(strcmp(a, "-da") == 0) {
				cparse.yydebug(true);
				debugaction = true;
				deco.ccout.debugtypes();
				cparse.debug(true);
				dumpscopes = true;
			}
			else if(strcmp(a, "-ddef") == 0)
				deco._notedef = true;
			else if(strcmp(a, "-ddecl") == 0)
				deco._notedecl = true;
			// -save-temps

			// Code Generation Options

			// Optimization Options
		
			else
				fprintf(stderr,
					"deco: unknown option `%s'\n", a);
			continue;
		}

		++nfiles;
		csz *fname = strcmp(a, "-") == 0 ? 0 : a;
		if(nolang) {
		    deco.parse_dcxx();
		    if(fname) {
			ParsePath p(fname);
			csz *ext = p.get_ext(".");
			if(ext) {
				if(strcmp(ext, "c") == 0) 
					deco.parse_c();
				else if(strcmp(ext, "cc") == 0
						|| strcmp(ext, "C") == 0
						|| strcmp(ext, "c++") == 0
						|| strcmp(ext, "cxx") == 0)
					deco.parse_cxx();
				else if(strcmp(ext, "dh") == 0)
					deco.parse_dhxx();
				else if(strcmp(ext, "hh") == 0
						|| strcmp(ext, "H") == 0)
					deco.parse_hxx();
				else if(strcmp(ext, "h") == 0)
					deco.parse_h();
			}
		    }
		}
		sz cppbuf[4096];
		if(cpp) 
			sprintf(cppbuf, "%s%s%s%s", cpp, 
				" -D__dcplusplus ",
				deco.hdr() ? " -D__dcplusplus_hdr" : "",
				cppopt);
		if(!deco.open_input(fname, cpp ? cppbuf : 0)) {
			perror(fname ? fname : "stdin" );
			break;
		}
		if(stop_cpp) {
			while(!lex.eof()) {
				while(!lex.is_empty()) {
					printf("%c", lex.get_char());
					lex.next_char();
				} 
				deco.get_input(false);
			}
			continue;
		}
		deco.predefine();
		if(!cparse.parse()) 
			deco.mprintf("", "cannot recover from previous errors");
		if(dumpscopes0)
			deco.dump_scopes();
		if(deco.errors() == 0 && deco.dcxx())
			deco.dcxx_to_cxx();
		if(deco.warnings() > 0)
			deco.mprintf("", "%d warning%s",
				deco.warnings(), deco.warnings() == 1 ? "" : "s");
		if(dumpscopes)
			deco.dump_scopes();
		if(deco.errors() > 0) {
			deco.mprintf("", "%d error%s",
				deco.errors(), deco.errors() == 1 ? "" : "s");
			rv = 1;
		} else if (!deco.nooutput()) {
			if(deco.hdr() || deco.ccout.interface())
				deco.ccout.outmask(
					(1<<Decl::OT_HDR)|(1<<Decl::OT_CODE));
			else
				deco.ccout.outmask((1<<Decl::OT_FROMHDR)|
					(1<<Decl::OT_HDR)|(1<<Decl::OT_CODE));
			deco.ccout.oprintf("// this file was auto-generated"
				" by deco " deco_VER_STR " from the command:");
			deco.ccout.onl();
			deco.ccout.oprintf("//\t");
			for(int i=0; i<argc; i++)
				deco.ccout.oprintf("%s ", argv[i]);
			deco.ccout.onl(2); 
			deco.print_syms();
		}
		deco.close_input();
	}
	if(!nfiles) {
		fprintf(stderr, "deco: nothing to do\n");
		exit(1);
	} 
	return rv;
}

