/*  $Id: YLex.hh,v 1.7 1999/07/16 20:18:52 shaggy Exp $
 *      Yacc Lexical Analysis
 *      relative path: util/shacc
 */ 

#include <stdx1.h>
#include <defines>
#include <Lexer>



// XXX - Need to add 'L' prefix for long characters (wchar_t)
//	Also need to set 'size' appropriately for number size
//	Need to detect negative numbers... or should that be handled
//	at a higher level?
//	Also need 'L' prefix for long strings,
//	and FfLl suffix for floats
//	For that matter, there is no support yet for floats
//	(optional eE suffix needed too, as well as sign)

class YLex : public Lexer {
	enum State {
		S_NEXT, S_SKIP, S_SKIP_NEXT, S_SPACE,
		S_NUM, S_ID,
		S_PERCENT, S_KW,
		S_CODE, S_CODE_PERCENT, S_CODE_DOLLAR, S_CODE_NUM,
		S_CODE_SCOMMENT, S_CODE_COMMENT, S_CODE_COMMENT2,
		S_CODE_SQUOTE, S_CODE_DQUOTE, S_CODE_QUOTE_ESC, 
		S_SQUOTE, S_QUOTE_ESC, S_QUOTE_OCT, S_QUOTE_HEX,
		S_SCOMMENT, S_COMMENT, S_COMMENT2, 
		S_EAT,
	} state;
	enum State old_state;
	char *str;		// string to store tokens in
	char *strp;		// current pointer in 'str'
	size_t str_size;	// total size of 'str'
	char *buf;		// buffer to store translations in
	char *bufp;		// current pointer in 'buf'
	size_t buf_size;	// total size of 'buf'
	char *name;		// buffer to store source name in
	char *namep;		// current pointer in 'name'
	size_t name_size;	// total size of 'name'
	uint32 num;		// numeric value of last token
	bool neg;		// last token is negative
	size_t size;		// size of last token
	int _blevel;		// brace level
	csz *_resultname;	// $$ = _resultname
	csz *_stackname;	// $x = _stackname[(x-1)+_stackofs]
	int _stackofs;
	int _startline;
public:
	enum Token {
		EOF_COMPLETE = 0,
		EOF_INCOMPLETE,
		NONE, INCOMPLETE, INVALID,

		ID, NUM, CHAR, CODE, ACTION,
		SECTION, COLON, OR, EOS,

		KW_TOKEN, KW_UNION, KW_START,
		KW_LEFT, KW_RIGHT, KW_NONASSOC, KW_PREC,
		KW_RPREC, KW_UNIQ,
		KW_PURE_PARSER,
		NUM_TOKENS,

		START_KW = KW_TOKEN,
		END_KW = KW_PURE_PARSER,
	};
private:
	Token _rtok;		// token to return (used for shared states)
	static csz *token_table[];
	Token token;
public:
	YLex(char *buf, size_t str_size);
	YLex& set_buf(char *ptr, size_t size)
		{ buf = ptr; buf_size = size; return *this; }
	YLex& set_tok(char *ptr, size_t size)
		{ str = ptr; str_size = size; return *this; }
	YLex& set_name(char *ptr, size_t size)
		{ name = ptr; name_size = size; return *this; }
	Token next_token();
	csz* token_string();
	csz* token_string(Token t) const;
	csz* token_string(Token t, bool tf) const;
	csz* token_string(Token t, uint32 val);
	csz* token_string(Token t, csz *str, size_t len = NPOS);
	Token last_token() const { return token; }
	uint32 last_num() const { return num; }
	uint last_char() const;
	csz* last_str() const { return str; }
	csz* last_id() const { return str; }
	size_t last_size() const { return size; }
	int last_startline() const { return _startline; }
	csz* source() const { return name; }
	void source(csz* str) { strcpy(name, str); }
	bool in_char() const { return state == S_SQUOTE; }
	void set_token(Token t) { token = t; }
	void set_resultname(csz *str) { _resultname = str; }
	void set_stackname(csz *str) { _stackname = str; }
	void set_stackofs(int ofs) { _stackofs = ofs; }
};

