/*  $Id: shacc.cc,v 1.35 2000/01/10 17:39:32 shaggy Exp $
 *      shacc: Shaggy's Homemade Alternative Compiler Compiler
 *      
 *      relative path: util/shacc
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define shacc_VER_MAJOR	0
#define shacc_VER_MINOR	6
#define shacc_VER_PATCH	0
#define shacc_VER_STR	"0.6"
#define shacc_VER_DATE	"January 10, 2000"
#define shacc_VER_NDATE	"01/10/00"

// #pragma Make Config_Needed += local
#pragma Make Includes_Banned += -nostdinc -nostdinc++
#pragma Make Flags_Banned += -static

#define DEBUG

#include <libc_posix.h>
#include <stdx1.h>
#include <defines>

// #include <Object>
#include <StrTab>
#include <SymTab>

#include <assert.h>

#include "YLex.hh"

#include "shacc.config.h"


static char rcsid[] __attribute__((unused))
	= "$Id: shacc.cc,v 1.35 2000/01/10 17:39:32 shaggy Exp $"
	  "$Release: " shacc_VER_STR " $";


char _strbuf[MAXSTRSIZE];
YLex _ylex(_strbuf, sizeof(_strbuf));
StrTab strtab;
SymTab symtab(&strtab);

struct Rule {
	uint16 sym;	// index to first symbol to match
	uint8  ns;	// number of symbols
	uint8  skip;	// number of rules to skip to find next (0 at end)
	uint8  srp;	// standard rule precedence
	uint8  rrp;	// reduction rule precedence
	bool   uniq;	// reducion must be unique

	csz   *action;	// action string, nil for none
	int    shared;	// action is referenced by others
	int16  red;	// non-terminal to reduce to
	int	line;// line that action code comes from
};

Rule rule[MAXRULES];
uint16 nonterm[MAXNONTERM];	// indexes to first rule for each nonterminal
int16 sym[MAXRULES*AVGSYMS];	// tokens/nonterms to match
int nrules = 0;
int nnonterms = 0;
int nsyms = 0;
int ntokens = 257;
int16 start = 0;

int _input_fd;
char *_buf;
int _bufsize;

FILE *fout;

csz *parser;

csz *_stype = 0;

vStrTab::Id tokenstr[MAXTOKENS];
vStrTab::Id nontermstr[MAXNONTERM];

int	tokenref[MAXTOKENS];
int	nontermref[MAXNONTERM];
int	nontermloop[MAXNONTERM];
int	nontermempty[MAXNONTERM];

uint8	tokenprec[MAXTOKENS];

#define T(t)	YLex::t

#ifdef NDEBUG
#	define assertfail(c) internal_error(__FILE__, __LINE__)
#else
#	define assertfail(c) _assertfail(#c, __FILE__, __LINE__)
#endif

int _num_errors = 0;
int _num_warnings = 0;

bool pureparse = true;
bool yaccmode;
bool debug = false;

static sz*
strdup(csz *s)
{
	sz *m = (sz*)malloc(strlen(s)+1);
	if(!m)
		return 0;
	return strcpy(m, s);
}

int
veprintf(csz *fmt, va_list args)
{
	return vfprintf(stderr, fmt, args);
}

void
abort_exit()
{
	exit(1);
}


int
eprintf(csz *fmt, ...)
{
	int n;
	va_list ap;
	va_start(ap, fmt);
	n = veprintf(fmt, ap);
	va_end(ap);
	return n;
}

void
vmprintf(csz *prefix, csz *fmt, va_list args)
{
	eprintf("%s:%d: %s", _ylex.source(), _ylex.line_num(), prefix);
	veprintf(fmt, args);
	switch(_ylex.last_token()) {
	case T(NONE):
		eprintf("\n");
		break;
	case T(EOF_COMPLETE):
	case T(EOF_INCOMPLETE):
		eprintf("; at end of file\n"); 
		break;
	case T(INVALID):
		eprintf("; at `%s'\n", _ylex.token_string());
		break;
	default:
		eprintf("; before `%s'\n", _ylex.token_string());
	}
	_ylex.set_token(T(NONE));
}


void
mprintf(csz *prefix, csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vmprintf(prefix, fmt, ap);
	va_end(ap);
}


void
error(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	_num_errors++;
	vmprintf("", fmt, ap);
	va_end(ap);
}


void
warning(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	_num_warnings++;
	vmprintf("warning: ", fmt, ap);
	va_end(ap);
}


void
abort(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vmprintf("abort: ", fmt, ap);
	va_end(ap);
	abort_exit();
}

void
internal_error_exit()
{
	mprintf("", "Please send a full bug report to `shaggy@csh.rit.edu'");
	exit(2);
}


void
internal_error(csz *file, int line, csz *module)
{
	if(module) 
		mprintf("Internal Error ", "in %s (%s:%d)",
			module, file, line);
	else
		mprintf("Internal Error ", "(%s:%d)", file, line);
	internal_error_exit();
}

void
_assertfail(csz *cond, csz *file, int line)
{
	mprintf("", "Assertion \"%s\" failed", cond);
	internal_error(file, line, 0);
}


bool
open_input(csz *fname)
{
	if(!fname) 
		_input_fd = 0;
	else 
		_input_fd = open(fname, O_RDONLY);
	return _input_fd >= 0;
}

void
close_input()
{
	if(_input_fd)
		close(_input_fd);
}

void
get_input(bool)
{
	
	int n = read(_input_fd, _buf, _bufsize);
	if(n <= 0)
		_ylex.set_eof();
	else
		_ylex.set_src(_buf, n);
}

YLex::Token
lex(int /* *val */ )
{
	YLex::Token t;
	for(;;) {
		t = _ylex.next_token();
		switch(t) {
		case T(NONE):
			get_input(false); continue;
		case T(INCOMPLETE):
			get_input(true); continue;
		case T(EOF_COMPLETE):
			return t;
		case T(EOF_INCOMPLETE):
			if(_ylex.in_char())
				error("unterminated character constant");
			else
				error("incomplete final token");
			return T(EOF_COMPLETE);
		case T(INVALID): 
			error("invalid character: %d", _ylex.get_char());
			continue;
		default:
			break;
		}
		return t;
	}
}

csz*
symname(int s)
{
	if(s > 0)
		return strtab.string(tokenstr[s]);
	else if(s < 0) 
		return strtab.string(nontermstr[-s]);
	else
		return "<EOF>";
}

void
print_line_directive(int n = 0, csz *s = 0)
{
	fprintf(fout, "#line %d \"%s\"\n", n > 0 ? n : _ylex.line_num(),
				s ? s :	_ylex.source());
}

int
writeactions(csz *prefix)
{
	int i;
	for(i=0; i<nrules; i++) {
		int s = rule[i].sym;
		int ns = rule[i].ns;
		csz *a = rule[i].action;
		fprintf(fout, a ? "%scase %d: " : "%s/* %d: ", prefix, i);
		fprintf(fout, a ? "/* %s:" : "%s:", symname(-rule[i].red));
		while(ns-- > 0) 
			fprintf(fout, " %s", symname(sym[s++]));
		fprintf(fout, " */\n");
		if(rule[i].shared) 
			fprintf(fout, "%syyrule%d: /* %d */\n", prefix, i,
				rule[i].shared);
		if(a) {
			print_line_directive(rule[i].line);
			fprintf(fout, "%s\t%s\n", prefix, a);
		}
		if(a)
			fprintf(fout, "%s\tbreak;\n", prefix);
	}
	return nrules*3;
}

int
writerules(csz *prefix)
{
	int i;
	for(i=0; i<nrules; i++) 
		fprintf(fout,
		  "%s/* %d */ {\t%d,\t%d,\t%d,\t%d,\t%d,\t%d,%d },\n",
			prefix, i,
			rule[i].sym, rule[i].ns, rule[i].skip,
				rule[i].srp, rule[i].rrp,
				(int)rule[i].uniq,
				rule[i].action != 0);
	return nrules;
}

int
writesymbols(csz *prefix)
{
	int i;
	for(i=0; i<nsyms; i++) {
		if((i%6) == 0)
			fprintf(fout, i == 0 ? "%s/* %d */ " : "\n%s/* %d */ ",
					prefix, i);
		fprintf(fout, "\t%d,", sym[i]);
	}
	fprintf(fout, "\n");
	return nsyms/6+1;
}

int
writeprecs(csz *prefix)
{
	int i;
	for(i=0; i<ntokens; i++) {
		if((i%10) == 0)
			fprintf(fout, i == 0 ? "%s/* %d */ " : "\n%s/* %d */ ",
				prefix, i);
		fprintf(fout, "%3d,", tokenprec[i]);
	}
	fprintf(fout, "\n");
	return ntokens/8+1;
}

int
writenonterms(csz *prefix)
{
	int i;
	for(i=0; i<nnonterms; i++) {
		if((i%6) == 0)
			fprintf(fout, i == 0 ? "%s/* %d */ " : "\n%s/* %d */ ",
					prefix, i);
		fprintf(fout, "\t%d,", nonterm[i]);
	}
	fprintf(fout, "\n");
	return nnonterms/6+1;
}

int
writenames(csz *prefix)
{
	int i;
	for(i=nnonterms-1; i; i--) 
		fprintf(fout, "%s/* %d (%d) */\t\"%s\",\n", prefix,
			-i, nontermref[i], strtab.string(nontermstr[i]));
	for(i=0; i<ntokens; i++) 
		fprintf(fout, "%s/* %d (%d) */\t\"%s\",\n", prefix,
			i, tokenref[i],
			tokenstr[i] ?  strtab.string(tokenstr[i]) : "");
	return nnonterms+ntokens;
}

int
printtokens(csz *prefix, bool defines)
{
	int i=0;
	// vStrTab::Id strid;
	for(i=0; i<ntokens; i++) {
		if(tokenstr[i]) {
			fprintf(fout, 
				i < 256 ? "%s/* %s = %d */\n"
				: defines ? "%s#define\t%s\t%d\n"
				: "%s%s = %d,\n", prefix,
					strtab.string(tokenstr[i]), i);
		}
	}
#ifdef ndef
	while((strid = symtab.index(i++))) {
		vSymTab::Val v;
		symtab.get(strid, v);
		int n = (int)v;
		if(n > 0) 
			fprintf(fout, 
				n < 256 ? "%s/* %s = %d */\n"
				: defines ? "%s#define\t%s\t%d\n"
				: "%s%s = %d,\n", prefix,
					strtab.string(strid), n);
	}
#endif
	fprintf(fout, defines ? "%s#define\t%s\t%d\n"
			: "%s%s = %d,\n", prefix, "YYLASTTOKEN", ntokens-1);
	return i;
}


csz *
smallesttype(int min, int max)
{
	if(min >= 0 && max <= 255)
		return "yyu8";
	if(min >= -128 && max <= 127)
		return "yys8";
	if(min >= 0 && max <= 65535)
		return "yyu16";
	if(min >= -32768 && max <= 32767)
		return "yys16";
	if(min >= 0)
		return "yyu32";
	else
		return "yys32";
}

void
writeparser(csz *fname)
{
	FILE *f = fopen(fname, "r");
	char buf[512];
	char *s;
	int iline = 0, oline = 0;

	if(yaccmode)
		fprintf(fout, "#define YYYACC\n");
	if(!pureparse)
		fprintf(fout, "#define YYIMPURE\n");
	if(debug)
		fprintf(fout, "#define YYDEBUG 1\n");

	fprintf(fout, "\ntypedef signed char yys8;\n");
	fprintf(fout, "typedef unsigned char yyu8;\n");
	fprintf(fout, "typedef short yys16;\n");
	fprintf(fout, "typedef unsigned short yyu16;\n");
	fprintf(fout, "typedef int yys32;\n");
	fprintf(fout, "typedef unsigned int yyu32;\n");

	if(_stype) {
		fprintf(fout, "\ntypedef union %s YYSTYPE;\n\n", _stype);
	} else {
		fprintf(fout, "\n#ifndef YYSTYPE\n");
		fprintf(fout, "#\tdefine\tYYSTYPE\tint\n");
		fprintf(fout, "#endif\n\n");
	}
	fprintf(fout, "typedef %s YYTTYPE;\n",
			smallesttype(-nnonterms, ntokens));
	fprintf(fout, "typedef %s YYRTYPE;\n", smallesttype(0, nrules));


	if(!f) {
		perror(fname);
		exit(1);
	}
	while(fgets(buf, sizeof buf, f)) {
		if(iline++ != oline) {
			print_line_directive(iline, fname);
			oline = iline-1;
		}
		if((s = strstr(buf, "$ACTIONS$"))) {
			*s = '\0';
			oline += writeactions(buf);
		} else if((s = strstr(buf, "$RULES$"))) {
			*s = '\0';
			oline += writerules(buf);
		} else if((s = strstr(buf, "$SYMBOLS$"))) {
			*s = '\0';
			oline += writesymbols(buf);
		} else if((s = strstr(buf, "$NONTERMS$"))) {
			*s = '\0';
			oline += writenonterms(buf);
		} else if((s = strstr(buf, "$NAMES$"))) {
			*s = '\0';
			if(debug)
				oline += writenames(buf);
		} else if((s = strstr(buf, "$LIMITS$"))) {
			*s = '\0';
			fprintf(fout, "%sYYNRULES = %d,\n", buf, nrules);
			fprintf(fout, "%sYYNNONTERMS = %d,\n", buf, nnonterms);
			fprintf(fout, "%sYYNTOKENS = %d,\n", buf, ntokens);
			fprintf(fout, "%sYYNSYMS = %d,\n", buf, nsyms);
			fprintf(fout, "%sYYINITDEPTH = %d,\n", buf,  1000);
			oline += 5;
		} else if((s = strstr(buf, "$TOKENENUMS$"))) {
			*s = '\0';
			oline += printtokens(buf, false);
		} else if((s = strstr(buf, "$TOKENDEFS$"))) {
			*s = '\0';
			oline += printtokens(buf, true);
		} else if((s = strstr(buf, "$TOKENPREC$"))) {
			*s = '\0';
			oline += writeprecs(buf);
		} else {
			fprintf(fout, "%s", buf);
			++oline;
		}
	}
}


int
newchartoken(int prec)
{
	char buf[20];
	csz *fmt;
	vStrTab::Id strid;
	vSymTab::Val symval;
	int c = _ylex.last_char();

	if(c >= 32 && c <= 126)
		fmt = "'%c'";
	else switch(c) {
	case '\a': fmt = "'\\\\a'"; break;
	case '\b': fmt = "'\\\\b'"; break;
	case '\177': fmt = "'\\\\d'"; break;
	case '\033': fmt = "'\\\\e'"; break;
	case '\f': fmt = "'\\\\f'"; break;
	case '\n': fmt = "'\\\\n'"; break; 
	case '\r': fmt = "'\\\\r'"; break;
	case '\t': fmt = "'\\\\t'"; break;
	case '\v': fmt = "'\\\\v'"; break;
	case '\\': fmt = "'\\\\\\\\'"; break;
	default:
		fmt = "'\\\\%03o'"; break;
	}
	sprintf(buf, fmt, c);
	strid = strtab.id(buf);
	if(!symtab.get(strid, symval)) {
		tokenstr[c] = strid;
		symtab.add(strid, (vSymTab::Val)c);
		tokenprec[c] = prec;
	}
	return c;
}

csz *
actionstr(csz *str)
{
	int i;
	if(!strstr(str, "__LINE__"))
	    for(i=0; i<nrules; i++)
		if(rule[i].action && strcmp(rule[i].action, str) == 0) {
			char b[20];
			sprintf(b, "goto yyrule%d;", i);
			rule[i].shared++;
			return strdup(b);
		}
	return strdup(str);
}

int
parse_section1(void)
{
	YLex::Token t;
	YLex::Token d = T(NONE);
	vStrTab::Id strid = 0;
	vSymTab::Val symval;
	int lprec = 0;	// level of precedence
	int cprec = 0;	// current precedence

	while((t = lex(0)) && t != T(SECTION)) {
		switch(t) {
		case T(CODE):
			print_line_directive();
			fprintf(fout, "%s\n", _ylex.last_str());
			continue;
		case T(KW_TOKEN):
			cprec = 0;
		case T(KW_START):
			d = t;
			continue;
		case T(KW_LEFT):
		case T(KW_RIGHT):
		case T(KW_NONASSOC):
			lprec += 4;
			cprec = lprec+(t == T(KW_RIGHT))-(t == T(KW_LEFT));
			d = t;
			continue;
		case T(ID):
			strid = strtab.id(_ylex.last_str());
			if(d == T(KW_START)) {
				if(start == 0) {
					symtab.add(strid, (vSymTab::Val)-1);
					nontermstr[1] = strid;
					start = 1;
					continue;
				}
				break;
			}
			if(!symtab.get(strid, symval)) {
				tokenstr[ntokens] = strid;
				symval = (vSymTab::Val)(ntokens++);
				symtab.add(strid, symval);
			}
			tokenprec[(int)symval] = cprec;
			if(d)
				continue;
			break;
		case T(CHAR):
			if(d && d != T(KW_START)) {
				newchartoken(cprec);
				continue;
			} 
			break;
		case T(NUM):
			if(strid) {
				int n = _ylex.last_num();
				tokenstr[n] = tokenstr[(int)symval];
				symtab.set(strid, (vSymTab::Val)n);
				continue;
			}
			break;
		case T(KW_PURE_PARSER):
			pureparse = true;
			continue;
		case T(KW_UNION):
			t = lex(0);
			if(t == T(ACTION)) {
				_stype = strdup(_ylex.last_str());
				continue;
			}
		default:
			break;
		}
		error("parse error");
	}
	return 1;
}

int
parse_section2(void)
{
	YLex::Token t;
	enum { S_NEXT, S_NEWNT, S_RULE, S_NEWRULE } state;
	int cnonterm = -1;
	int crule = -1;
	bool newnt = false;
	bool bprec = false;
	bool brprec = false;
	vStrTab::Id strid;
	vSymTab::Val symval;
	int rprec = 0;
	int sprec = 0;

	state = S_NEXT;
	nonterm[0] = 0;
	rule[0].ns = 2;
	rule[0].sym = 0;
	rule[0].skip = 0;
	rule[0].action = 0;
	sym[0] = -1;
	sym[1] = 0;
	nnonterms = 1+start;
	nrules = 1;
	nsyms = 2;

	while((t = lex(0)) && t != T(SECTION)) {
		if(cnonterm >= 0 && !bprec && !brprec && (t == T(ID) || t == T(CHAR)
				|| t == T(ACTION) || t == T(OR) || t == T(EOS))) {
			if(crule < 0) {
				crule = nrules++;
				rule[crule].ns = 0;
				rule[crule].sym = nsyms;
				rule[crule].skip = 0;
				rule[crule].action = 0;
				rule[crule].shared = 0;
				rule[crule].red = cnonterm;
				rule[crule].uniq = 0;
				rprec = 0;
				sprec = 0;
			} else if(rule[crule].action && t != T(OR)
						&& t != T(EOS)) {
				char buf[10];
				rule[nrules].action = rule[crule].action;
				rule[nrules].line = rule[crule].line;
				rule[crule].action = 0;
				rule[nrules].ns = 0;
				rule[nrules].sym = 0;
				rule[nrules].skip = 0;
				rule[nrules].red = nnonterms;
				sym[nsyms++] = -nnonterms;
				sprintf(buf, "<%d>", nrules);
				nontermstr[nnonterms] = strtab.id(buf);
				nonterm[nnonterms++] = nrules++;
				rule[crule].ns++;
			}
		}
		switch(t) {
		case T(CODE):
			print_line_directive();
			fprintf(fout, "%s\n", _ylex.last_str());
			continue;
		case T(ID):
		case T(CHAR):
			if(newnt)
				break;
			if(t == T(ID)) {
				strid = strtab.id(_ylex.last_str());
				if(!symtab.get(strid, symval)) {
					nontermstr[nnonterms] = strid;
					nonterm[nnonterms] = 0;
					symval = (vSymTab::Val)(-(nnonterms++));
					symtab.add(strid, symval);
				}
			} else 
				symval = (vSymTab::Val)newchartoken(0);
			if(cnonterm < 0) {
				newnt = true;
				continue;
			}
			if((int)symval >=  0) {
				int p = tokenprec[(int)symval];
				if(p != 0)
					sprec = (p+1)&~3;
			}
			if(bprec) {
				if((int)symval < 0)
					break;
				bprec = false;
				continue;
			}
			_ylex.set_stackofs(-(++rule[crule].ns));
			sym[nsyms++] = (int)symval;
			continue;
		case T(NUM):
			if(brprec) {
				rprec = _ylex.last_num();
				brprec = false;
				continue;
			}
			break;
		case T(COLON):
			if(newnt) {
				newnt = false;
				if((cnonterm = -(int)symval) <= 0) {
					error("using token on lhs");
					break;
				}
				crule = -1;
				if(nonterm[cnonterm])
					error("redefining %s", strtab.string(
						nontermstr[cnonterm]));
				nonterm[cnonterm] = nrules;
				_ylex.set_stackofs(0);
				continue;
			}	
			break;
		case T(KW_PREC):
			if(cnonterm > 0 && !bprec) {
				bprec = true;
				continue;
			}
			break;
		case T(KW_RPREC):
			if(cnonterm > 0 && !brprec) {
				brprec = true;
				continue;
			}
			break;
		case T(KW_UNIQ):
			if(cnonterm > 0) {
				rule[crule].uniq = true;
				continue;
			}
			break;
		case T(ACTION):
			if(cnonterm > 0) {
				rule[crule].action =
						 actionstr(_ylex.last_str());
				rule[crule].skip = nrules-crule;
				rule[crule].line = _ylex.last_startline();
				rule[crule].srp = sprec;
				rule[crule].rrp = rprec;
				continue;
			}
			break;
		case T(OR):
			if(cnonterm > 0) {
				rule[crule].skip = nrules-crule;
				rule[crule].srp = sprec;
				rule[crule].rrp = rprec;
				crule = -1;
				_ylex.set_stackofs(0);
				continue;
			}
			break;
		case T(EOS):
			if(cnonterm > 0) {
				rule[crule].skip = 0;
				rule[crule].srp = sprec;
				rule[crule].rrp = rprec;
			}
			cnonterm = -1;
			continue;
			break;
		default:
			break;
		}	
		error("parse error");	
	}
	return 1;
}

int
parse_section3(void)
{
	print_line_directive();
	while(!_ylex.eof()) {
		if(!_ylex.next_char()) {
			get_input(0);
			continue;
		}
		fprintf(fout, "%c", _ylex.get_char());
	}
	return 1;
}

int
canbeempty(int nt)
{
	int y = nontermempty[nt];
	int b;
	if(y != -1) 
		return y;
	y = nonterm[nt];
	b = 0;
	nontermempty[nt] = -2;
	while(y > 0) {
		int i, b = 0;
		for(i=0; i<rule[y].ns; i++) {
			int m = sym[rule[y].sym+i];
			if(m >= 0)
				goto notallempty;
			m = canbeempty(-m);
			if(m < 0) {
				b = -2;
				goto notallempty;
			}
			if(m == 0)
				goto notallempty;
		}
		nontermempty[nt] = 1;
		return 1;
	notallempty:
		if(rule[y].skip)
			y += rule[y].skip;
		else
			y = -1;
	}
	if(b == 0)
		nontermempty[nt] = b;
	return b;
}

int
loopcheck(int nt)
{
	int y = nontermloop[nt];
	if(y < 0)
		return 0;
	else if(y > 0) {
		char buf[1000];
		char *bp = buf;
		y = nt;
		bp +=  sprintf(bp, "infinite recursion: %s",
				strtab.string(nontermstr[y]));
		do {
			y = nontermloop[y];
			bp += sprintf(bp, " -> %s",
					strtab.string(nontermstr[y]));
		} while(y != nt);
		error(buf);
		return 1;
	}
	y = nonterm[nt];
	while(y > 0) {
		int i, n = 0;
		for(i=0; i<rule[y].ns; i++) {
			int m = sym[rule[y].sym+i];
			if(m < 0) {
				m = -m;
				if(!canbeempty(m)) {
					if(n)
						goto nextrule;
					n = m;
				}
			} else
				goto nextrule; 
		}
		nontermloop[nt] = n;
		i = loopcheck(n);
		nontermloop[nt] = 0;
		if(i)
			for(i=0; i<rule[y].ns; i++) {
				int m = -sym[rule[y].sym+i];
				nontermloop[nt] = m;
				loopcheck(m);
				nontermloop[nt] = 0;
			}
	nextrule:
		if(rule[y].skip)
			y += rule[y].skip;
		else
			y = -1;
	}
	nontermloop[nt] = -1;
	return 0;
}

void
follow(int r)
{
	int i;
	for(i=0; i<rule[r].ns; i++) {
		int s = sym[rule[r].sym + i];
		if(s > 0)
			tokenref[s]++;
		else if(s < 0) {
			int y;
			s = -s;
			if(!nontermref[s]++) {
				loopcheck(s);
				y = nonterm[s];
				while(y > 0) {
					follow(y);
					if(rule[y].skip)
						y += rule[y].skip;
					else
						y = -1;
				}
			}
		}
	}
}

int
parse(void)
{
	int i;
	int utoken = 0, urule = 0, unonterm = 0;

	_ylex.set_resultname("yyval");
	_ylex.set_stackname("yyvsp");
	_ylex.set_stackofs(0);

	symtab.add((tokenstr[0] = strtab.id("$")), (vSymTab::Val)0);
	symtab.add((tokenstr[1] = strtab.id("error")), (vSymTab::Val)1);
	symtab.add((tokenstr[2] = strtab.id("?")), (vSymTab::Val)2);

	if(!parse_section1())
		return 0;
	if(!parse_section2())
		return 0;

	for(i=1; i<nnonterms; i++) {
		nontermref[i] = 0;
		nontermloop[i] = 0;
		nontermempty[i] = -1;
	}
	for(i=1; i<ntokens; i++)
		tokenref[i] = 0;

	follow(0);

	for(i=1; i<nnonterms; i++) {
		if(!nontermref[i]) {
			int y = nonterm[i];
			while(y > 0) {
				++urule;
				if(rule[y].skip)
					y += rule[y].skip;
				else
					y = -1;
			}
			++unonterm;
		}
		else if(!nonterm[i])
			error("%s undefined", strtab.string(nontermstr[i]));
	}
	for(i=3; i<ntokens; i++)
		if(tokenstr[i] && !tokenref[i])
			++utoken;

	if(utoken)
		warning("%d unreferenced tokens", utoken);
	if(unonterm)
		warning("%d unreferenced non-terminals", unonterm);
	if(urule)
		warning("%d unreferenced rules", urule);
	writeparser(parser);
	if(!parse_section3())
		return 0;
	return 1;
}



int
main(int argc, sz **argv)
{
	char bufbuf[4096];
	char namebuf[1024];
	bool dumptokens = false;
	bool defines = false;
	char *fprefix = 0;
	char fprefixbuf[1024];
	char *outfile = 0;
	char *houtfile = 0;
	char buf[1024];
	char *ext = ".c";

	yaccmode = (strcmp(argv[0], "yacc") == 0);
	_buf = bufbuf;	
	_bufsize = sizeof(bufbuf);
	_ylex.set_buf(bufbuf, sizeof(bufbuf));
	_ylex.set_name(namebuf, sizeof(namebuf));

	int argnum = 0;
	parser = 0;
	while(++argnum < argc) {
		if(!strcmp(argv[argnum], "-b")) {
			fprefix = argv[++argnum];
		} else if(!strcmp(argv[argnum], "-d")) {
			defines = true;
			dumptokens = true;
		} else if(!strcmp(argv[argnum], "-e")) {
			dumptokens = true;
		} else if(!strcmp(argv[argnum], "-t")) {
			debug = true;
		} else if(!strcmp(argv[argnum], "-i")){
			pureparse = false;
		} else if(!strcmp(argv[argnum], "-y")) {
			yaccmode = true;
			pureparse = false;
		} else if(!strcmp(argv[argnum], "-o")) {
			outfile = argv[++argnum];
		} else if(!strcmp(argv[argnum], "-h")) {
			houtfile = argv[++argnum];
		} else if(!strcmp(argv[argnum], "-p")) {
			parser = argv[++argnum];
		} else if(!strcmp(argv[argnum], "-V")) {
			printf("shacc " shacc_VER_STR " ("
				shacc_VER_DATE ")\n");
			printf("CPARSER=%s\nCCPARSER=%s\n",
				CPARSER, CCPARSER); 
		} else
			break;
	}
	if((argc-argnum) != 1) {
		eprintf("usage: shacc [-detyiV] [-b fileprefix] [-o outfile] "
			"[-h hdrfile] [-p parser] filename\n");
		exit(1);
	}
	if(!open_input(argv[argnum])) {
		perror(argv[argnum]);
		exit(1);
	}
	strcpy(namebuf, argv[argnum]);
	if(!yaccmode)
		pureparse = true;
	if(outfile)
		ext = strrchr(outfile, '.');
	if(!fprefix) {
		if(yaccmode) 
			fprefix = "y";
		else {
			char *c;
			strcpy(fprefixbuf, outfile ? outfile : namebuf);
			c = strrchr(fprefixbuf, '.');
			if(c)
				*c = '\0';
			fprefix = fprefixbuf;
		}
	}
	if(outfile)
		strcpy(buf, outfile);
	else
		sprintf(buf, "%s.tab%s", fprefix, ext);
	if(!parser) {
		if(strcmp(ext, ".cc") == 0)
			parser = CCPARSER;
		else
			parser = CPARSER;
	}
	fout = fopen(buf, "w");
	fprintf(fout, "/* this file was auto-generated by shacc " shacc_VER_STR
				" from the command:\n");
	fprintf(fout, " *\t");
	for(int i=0; i<argc; i++)
		fprintf(fout, "%s ", argv[i]);
	fprintf(fout, "\n */\n");
	if(!parse()) 
		eprintf("cannot recover from previous errors\n");
	fclose(fout);
	if(dumptokens) {
		if(houtfile)
			strcpy(buf, houtfile);
		else
			sprintf(buf, "%s.tab.h", fprefix);
		fout = tmpfile();
		fprintf(fout, "/* this header file was auto-generated by "
				"shacc " shacc_VER_STR
				" from the command:\n");
		fprintf(fout, " *\t");
		for(int i=0; i<argc; i++)
			fprintf(fout, "%s ", argv[i]);
		fprintf(fout, "\n */\n");

		printtokens("", defines);

		/* this is far from efficient, but it's fairly simple,
		 * and doesn't usually have to work on big files
		 */
		FILE *ftmp = fout;	
		rewind(ftmp);
		FILE *fin = fopen(buf, "r");
		bool match = false;
		int c;
		if(fin) {
			while((c = fgetc(fin)) == fgetc(fout))
				if(c == EOF) {
					match = true;
					break;
				}
			fclose(fin);
		}
		if(match)
			printf("%s: unchanged\n", buf);
		else {
			fout = fopen(buf, "w");
			rewind(ftmp);
			while((c = fgetc(ftmp)) && c != EOF)
				fputc(c, fout);
			fclose(fout);
		}
		fclose(ftmp);
	}
	close_input();
	return _num_errors > 0 ? 1 : 0;
}

